# Création du CFG

Pour créer le CFG, on utilise une variable data qui contient des informations administratives comme
les variables locales et leur position dans la pile. Pour transformer une expression, on procède
par le bas, pendant la transformation d'une instruction, on s'occupe d'abord de traiter les
instructions qui suivent. Et on utilise data.next qui représente le bloc en cours ou un label avec
qui contient les instructions suivantes, pour lier l'instruction courante avec la suite: la fonction
`add_data_ins` s'occupe de rajouter une instruction non finale. Il y a aussi des fonctions qui
permettent de forcer la création d'un bloc comme `force_next_label` utilisée pour la transformation
d'un while, ou encore `prefers_next_label` qui ferme et transforme un bloc non vide en label, elle
est utilisé pour les branchements conditionnels (CIF, EIF) pour éviter une duplication du code :
pour traiter :

```c
iDEB;
if (e) {
    iA;
} else {
    iB;
}
iFIN;
```

sans cette fonction, on générerait :

        +-> iA iFIN
iDEB e -+
        +-> iB iFIN

que l'on ne peut pas se permettre si iFIN est très très grand (mais si iFIN est juste un JmpC, on
peut se le permettre). On est donc obligé de créer un nouveau label pour iFIN ce qui donne :

        +-> iA -+
iDEB e -+       +-> iFIN
        +-> iB -+

### While

Pour éviter de rajouter des jumps non nécessaires :
```c
iDEB;
while (iCOND) {
    iA;
}
iFIN;
```
sera traité en
                   +---------<--------+
iDEB iCOND JmpC -+-+-> iA iCOND JmpC -+-+-> iFIN
                 |                      |
                 +----------->----------+

### Call

Lors d'un appel de fonction pour l'instant, j'ai utilisé l'instruction `Pop r` pour récupérer le
résultat dans le registre r. Peut être que plus tard je le remplacerait par un `Mov` depuis le
registre utilisé pour le retour des fonctions plutôt que d'empiler sur la pile.

# Analyse du CFG

## Caractéristiques communes

Pour les analyse chaque fonction est traitée indépendamment. Et on va parcourir le graphe pour
analyser chaque bloc un après l'autre. Les analyses se passent en deux phases, une phase d'analyse
et une phase de mise à jour.

### Phase de mise à jour

Pendant cette phase, le CFG n'est pas modifié, à chaque bloc, on va associer un dictionnaire qui à
chaque variable et registre (`type position = VLoc of location | VReg of register`) va associer une
valeur (constante, nécessité, déclaration), ainsi que des données administratives. C'est à dire

BLOC1 :

  defs:
    Reg r1 -> 2
    Reg r2 -> 5
    Reg r3 -> None
    Loc (Local 0) -> Unsafe
    …

… (valeurs administratives)

BLOC2 :

  defs:
    Reg r25 -> 2
    Loc r42 -> 5
    …

… (valeurs administratives)

…

On ne sauvegarde pas les définitions pour chaque ligne, seulement pour chaque bloc, et lors d'une
passe, on peut redériver les changements à partir des instructions.

Dans le code, la table defs est le type `def_tbl` (ou `req_tbl`) i.e. une `value LocMap.t`.
La table des blocs, est de type `blkdef_tbl` (ou blkreq_tbl) i.e. une `blk_data BlockMap.t` où
`blk_data` contient une `def_tbl`. Par exemple dans le cas de l'analyse des constantes on a :

```ocaml
(* Pour le détail de def_tbl, cf Spécificités de l'analyse des constantes *)
type def_tbl = { unsafe_vars : bool LocMap.t; defs : value LocMap.t }

(* Une définition de bloc vide *)
let empty_defs = { unsafe_vars = LocMap.empty; defs = LocMap.empty }

(* Les données d'un bloc *)
type blk_data = {
  (* Définitions au début du bloc *)
  init_def : def_tbl option; (* le fait que ce soit une option est spécifique à l'analyse des constantes*)
  (* Position dans le tri topologique *)
  pos : int;
  (* Label suivant dans le tri topologique (où None si c'est la fin)*)
  next_topo_label : label option;
  (* Renvoie si ce bloc est atteignable *)
  reachable : bool;
}

(* Table qui à chaque bloc associe ses données *)
type blkdef_tbl = blk_data BlockMap.t
```

Pour analyser une fonction :

 - on commence par trier "topologiquement" (modulo les boucles while) les label des blocs.
 - on traite le premier bloc (cf spécificités)
    - on regarde toutes les instructions et on dérive la table de définition
    - on transfère les informations dans les blocs suivant en les fusionnant avec les informations
      que les blocs suivants connaissent déjà de leurs variables
    - si c'est un bloc déjà traité (boucle while) on recommence à traiter les blocs à partir de
      celui là
    - sinon, on passe au bloc suivant dans le tri topologique

### Phase de mise à jour

Une fois les tables de définitions de début de bloc obtenues pour chaque bloc, on fait une deuxième
passe sur chaque bloc dans laquelle on modifie chaque instruction (suppression, utilisation des
constantes…), on redérive à partir de la liste des instruction la table pour les lignes
intermédiaires.

### Autres remarques

Pendant les analyses, il est important de différencier les variables locales, or avec seulement la
position dans la pile on peut avoir des problèmes avec des sous blocs (notamment pour l'analyse
d'initialisation) :
```c
int main() {
    int a;
    {
        int b: // Position 1 dans la pile
        b = 2;
        truc(b);
    }
    {
        int c // Même position mais c'est une autre variable !
        truc(c);
    }
}
```

Pour régler ce problème, chaque variable locale a aussi un identifiant unique (dans la fonction).

## Spécificités de l'analyse des constantes

L'analyse se fait du début vers la fin du programme. Le type `value` de cette analyse est :
`type value = VNull | VInt of int | VLoc of location` donc si `Reg r` est associé à `VInt 2`, cela
veut dire que sur toutes les exécutions, `Reg r` a une valeur constante qui est l'entier 2.
Plusieurs valeurs possibles est dénoté par une absence d'affectation.

On a besoin de différentier un bloc qui a des variables avec plusieurs valeurs possibles et un bloc
qui n'a jamais été vu, pour ce faire, dans `blk_data` on a `init_def : def_tbl option`, où `None`
signifie que l'on a jamais vu le bloc, et `Some empty` que toutes les variables ont plusieurs
valeurs possibles

Il faut être précautionneux avec les variables qui perdent leur adresse par exemple :

```c
int evil(int *x) {
    *x = 1;
}

int main() {
    int x;
    x = 0;
    evil(&x);
    if (x) return 0;
    else return 1;
}
```

Pour cela, quand une variable perd son adresse elle est marqué comme unsafe (les variables globales
sont implicitement unsafe). Lorsque l'on fait une opération dangereuse (appel de fonction, StoreI
vers une destination inconnue), on enlève toutes les affectations (i.e. valeur inconnue).

## Spécificités de l'analyse du code mort

Pour le code mort, on procède de manière analogue à l'analyse des constantes, sauf que à la place
de faire du début du code vers la fin, on va procéder de la fin vers le début. Quand on utilise une
variable où un registre on marque cette variable comme `VNeeded`, puis lorsque que l'on a une
instruction qui affecte une valeur, on enlève le `VNeeded`.

Ces informations servent lors de la phase de mise à jour à enlever les instructions qui n'affecte
pas des variables dont on a besoin plus tard.

On a les mêmes difficultés avec les variables qui perdent leur adresse dont on ne peut pas prédire
si elles sont utilisées ou pas dans la suite du programme. Pour cela on les marque comme étant
`VUnsafe` et on présume qu'elle est requise et donc on ne supprime jamais des instructions qui 
l'utilise (dans le doute).

## Spécificités de l'analyse des initialisation

On procède du début vers la fin du code. On a pas besoin de traiter les boucles puisque l'on ne peut
pas déinitialiser une variable, on a pas besoin d’itérer les boucles puisqu'on aura déjà repéré les
variables non initialisées.

