open Cfg
open PrettyPrinter
open Analysis

(* Pour les types utilisés, de plus d'amples explications sont écrites dans la doc *)
module BlockMap = Map.Make (struct
  type t = label

  let compare = compare
end)

(* Plusieurs valeurs possibles sont représentés par une absence d'affectation *)
type value = VNull | VInt of int | VLoc of location

(* Formate une valeur *)
let pp_value p =
  let open Format in
  function
  | None -> fprintf p "unk"
  | Some VNull -> fprintf p "nil"
  | Some (VInt i) -> fprintf p "%d" i
  | Some (VLoc l) -> fprintf p "%a" pp_location l

(*
 * Table des définitions (stocké pour chaque bloc)
 *
 * unsafe_vars contient les variable dont l'utilisateur connaît leur adresse,
 * On suppose que l'utilisateur ne doit pas essayer de deviner l'adresse d'une 
 * variable à partir d'une autre adresse. (Si ça marche pas c'est sa faute sinon).
 * Il ne peut pas essayer de deviner l'adresse avec un programme du style :
 * ```
 * int x;
 * int y;
 * int *z;
 * z = &x + 1;
 * ``` *)
type def_tbl = { unsafe_vars : bool LocMap.t; defs : value LocMap.t }

(* Une définition de bloc vide *)
let empty_defs = { unsafe_vars = LocMap.empty; defs = LocMap.empty }

(* Les données d'un bloc *)
type blk_data = {
  (* Définitions au début du bloc *)
  init_def : def_tbl option;
  (* Position dans le tri topologique *)
  pos : int;
  (* Label suivant dans le tri topologique (où None si c'est la fin)*)
  next_topo_label : label option;
  (* Renvoie si ce bloc est atteignable *)
  reachable : bool;
}

(* Table qui à chaque bloc associe ses données *)
type blkdef_tbl = blk_data BlockMap.t

(* Affecte une valeur constante à la position p *)
let set_val (d : def_tbl) (p : position) (v : value option) =
  (* Format.printf "Set %a: %a\n" pp_position p pp_value v; *)
  match v with
  | None -> { d with defs = LocMap.remove p d.defs }
  | Some v -> { d with defs = LocMap.add p v d.defs }

(* Renvoie la valeur potentiellement constante de la position p *)
let get_val (t : def_tbl) (p : position) =
  let res = LocMap.find_opt p t.defs in
  (* Format.printf "Get %a: %a\n" pp_position p pp_value res; *)
  res

(* Récupère une valeur constante booléenne *)
let get_val_bool (t : def_tbl) (p : position) =
  match get_val t p with
  | Some (VInt 0) | Some VNull -> Some false
  | Some _ -> Some true
  | None -> None

(* Récupère une valeur à l'adresse stocké dans la valeur potentiellement constante v
   utile pour les *x = e *)
let get_val_at_addr (t : def_tbl) (v : value option) =
  match v with Some (VLoc l) -> get_val t (Loc l) | _ -> None

(* Marque une variable comme unsafe, c'est à dire qu'elle a donné son adresse *)
let set_unsafe (p : position) (d : def_tbl) =
  { d with unsafe_vars = LocMap.add p true d.unsafe_vars }

(* Fonction d'initialisation des blocs de données blkdata *)
let rec init_blkdefs (l : label list) (pos : int) (blkdefs : blkdef_tbl) =
  match l with
  | [] -> blkdefs
  | [ x ] ->
      BlockMap.add x
        { init_def = None; pos; next_topo_label = None; reachable = false }
        blkdefs
  | x :: y :: q ->
      BlockMap.add x
        { init_def = None; pos; next_topo_label = Some y; reachable = false }
        blkdefs
      |> init_blkdefs (y :: q) (pos + 1)

(* Fusionne le bloc de donnée a avec les définitions b venant d'un prédécesseur  *)
let merge_defs (a : blk_data) (b : def_tbl) : bool * blk_data =
  let changed = ref false in
  let union_def p x y =
    match (x, y) with
    | Some a, Some b when a = b -> Some a
    | a, _ ->
        changed := !changed || a <> None;
        None
  in
  let union_unsafe x y =
    match (x, y) with
    | Some true, _ -> Some true
    | _, Some true ->
        changed := true;
        Some true
    | _, _ -> Some false
  in
  let defs_res =
    match a.init_def with
    | None ->
        (* C'est la première fois que l'on croise le bloc*)
        changed :=
          (not (LocMap.is_empty b.defs)) || not (LocMap.is_empty b.unsafe_vars);
        b
    | Some tb ->
        (* Si on la déjà croisé, on fusionne ses définitions et ses déclarations unsafe *)
        {
          defs = LocMap.merge union_def tb.defs b.defs;
          unsafe_vars =
            LocMap.merge (Fun.const union_unsafe) tb.unsafe_vars b.unsafe_vars;
        }
  in
  (* On renvoie aussi si on a eu des nouvelles informations *)
  (!changed, { a with init_def = Some defs_res })

(* On calcule si possible une opération unaire de constante *)
let analyse_monop (defs : def_tbl) (o : monop) (v : value option) =
  match (o, v) with
  | ADDI a, Some (VInt i) -> Some (VInt (i + a))
  | MOV, v -> v
  | DEREF, Some (VLoc l) -> get_val defs (Loc l)
  | MINUS, Some (VInt i) -> Some (VInt (-i))
  | NOT, Some (VInt i) -> Some (VInt (Int.lognot i))
  | _ -> None

(* On calcule si possible une opération binaire de constantes, on ne s'occupe pas des pointeurs
   car on ne peut pas faire de calcul intéressant sans que ça devienne UndefineBehavior*)
let analyse_binnop (defs : def_tbl) (o : binop) (v1 : value option)
    (v2 : value option) =
  match (o, v1, v2) with
  | ADD, Some (VInt i), Some (VInt j) -> Some (VInt (i + j))
  | SUB, Some (VInt i), Some (VInt j) -> Some (VInt (i - j))
  | MUL, Some (VInt i), Some (VInt j) -> Some (VInt (i * j))
  | DIV, Some (VInt i), Some (VInt j) -> Some (VInt (i / j))
  | MOD, Some (VInt i), Some (VInt j) -> Some (VInt (i mod abs j))
  | CMP_EQ, Some (VInt i), Some (VInt j) -> Some (VInt (if i = j then 1 else 0))
  | CMP_LT, Some (VInt i), Some (VInt j) -> Some (VInt (if i < j then 1 else 0))
  | CMP_LE, Some (VInt i), Some (VInt j) ->
      Some (VInt (if i <= j then 1 else 0))
  | _ -> None

(* On analyse le programme… *)
let analyse (program : program) =
  (* On vient de faire une opération dont on ne peut pas garantir la destination, on oublie
     toutes les variables qui ont donné leur adresse et les variables globales cf set_val_at_addr *)
  let forget_unsafe_defs (t : def_tbl) =
    {
      t with
      defs =
        LocMap.fold
          (fun p u t -> if u then LocMap.remove p t else t)
          t.unsafe_vars t.defs
        |> Fun.flip
             (List.fold_left (fun t n -> LocMap.remove (Loc (Global n)) t))
             program.globals;
    }
  in
  (* On veut affecter la variable à l'adresse vpos à v *)
  let set_val_at_addr (d : def_tbl) (vpos : value option) (v : value option) =
    match vpos with
    | Some (VLoc l) -> set_val d (Loc l) v
    | a ->
        if a <> None then
          Format.eprintf "warning: cf: Possible erreur de segmentation";
        (* Impossible de savoir où on set, on doit supprime toutes les variables dont on a
           récupéré l'adresse *)
        forget_unsafe_defs d
  in
  (* On analyse une instruction *)
  let analyse_ins (defs : def_tbl) (i : non_final_instruction) =
    match i with
    | Cst (r, i) -> set_val defs (Reg r) (Some (VInt i))
    | Store (l, r) -> set_val defs (Loc l) (get_val defs (Reg r))
    | StoreI (l, r) ->
        set_val_at_addr defs (get_val defs (Loc l)) (get_val defs (Reg r))
    | StoreR (raddr, r) ->
        set_val_at_addr defs (get_val defs (Reg raddr)) (get_val defs (Reg r))
    | Load (r, l) -> set_val defs (Reg r) (get_val defs (Loc l))
    | Addr (r, l) -> set_val defs (Reg r) (Some (VLoc l)) |> set_unsafe (Loc l)
    | Monop (res, op, r) ->
        get_val defs (Reg r) |> analyse_monop defs op |> set_val defs (Reg res)
    | Binop (res, op, r1, r2) ->
        analyse_binnop defs op (get_val defs (Reg r1)) (get_val defs (Reg r2))
        |> set_val defs (Reg res)
    | Call _ | CallR _ -> forget_unsafe_defs defs
    | PushCallArg _ | PopCallRes _ | Nop | PrintString _ | FreeRegs _ -> defs
  in
  (* On traite le programme et on récupère les informations sur chaque bloc *)
  let blocs =
    (* map sur chaque fonction de program *)
    List.map
      (fun (f : function_cfg) ->
        (* On récupère l'ordre dans lequel on va traiter les blocs *)
        let topo = topo_sort f.entry f.graph in
        (* Fonction auxiliaire d'analyse du bloc l *)
        let rec analyse_blk (l : label) (blkdefs : blkdef_tbl) =
          (* On récupère ses données *)
          let data = BlockMap.find l blkdefs in
          (* Quand on a terminé de traiter le bloc l, on va mettre à jour les définitions des blocs
             incidents, cette fonction s'occupe de mettre à jour jlabel avec les définitions defs
             obtenues avec l'analyse du bloc courant, next_label sert à revenir en arrière au besoin *)
          let update_defs (defs : def_tbl) (jlabel : label)
              (next_label : label option ref) (blkdefs : blkdef_tbl) :
              blkdef_tbl =
            BlockMap.update jlabel
              (fun (jump_data : blk_data option) ->
                let jump_data = Option.get jump_data in
                let changed, new_j_defs = merge_defs jump_data defs in
                (* Si on a rajouter des informations sur un bloc précédent, on demande de retraiter
                   les blocs à partir de celui ci*)
                (* On peut écraser sans crainte car la seule façon d'avoir une boucle et sur
                   un while qui génère un JmpC avec un label qui boucle et un qui ne boucle pas. *)
                if changed && jump_data.pos <= data.pos then
                  next_label := Some jlabel;
                Some { new_j_defs with reachable = true })
              blkdefs
          in
          (* On récupère les instructions du blocs *)
          let ins, fins = get_block l f.graph in
          (* On traite le bloc en fonction de l'instruction finale, et on demande quel bloc on
             doit exécuter après *)
          let next_label, blkdefs =
            match (fins, data.reachable) with
            (* Si le bloc n'est pas atteignable ou si on a un Return comme on n'a besoin que des
               définitions au début du bloc, on n'a pas besoin de traiter ces cas ici*)
            | _, false | Return _, true -> (data.next_topo_label, blkdefs)
            | Jmp j, true -> begin
                (* On analyse le bloc et on transmet les constantes au bloc j *)
                let defs =
                  List.fold_left analyse_ins
                    (Option.value data.init_def ~default:empty_defs)
                    ins
                in
                let next_label = ref data.next_topo_label in
                let blkdefs = update_defs defs j next_label blkdefs in
                (!next_label, blkdefs)
              end
            | JmpC (r, j1, j2), true ->
                (* On analyse le bloc et on transmet les constantes au bloc j1 et j2 *)
                let defs =
                  List.fold_left analyse_ins
                    (Option.value data.init_def ~default:empty_defs)
                    ins
                in
                let next_label = ref data.next_topo_label in
                (* On essaye de voir si c'est un JmpC constant *)
                let blkdefs =
                  match get_val_bool defs (Reg r) with
                  | Some true -> update_defs defs j1 next_label blkdefs
                  | Some false -> update_defs defs j2 next_label blkdefs
                  | None ->
                      update_defs defs j1 next_label blkdefs
                      |> update_defs defs j2 next_label
                in
                (!next_label, blkdefs)
          in
          (* On a terminé ce bloc là (pour l'instant), on saute au bloc suivant *)
          match next_label with
          | None -> blkdefs
          | Some l -> analyse_blk l blkdefs
        in
        (* On traire chaque bloc de la fonction courante *)
        let blksdefs =
          init_blkdefs topo 0 BlockMap.empty
          |> BlockMap.update f.entry (fun jl ->
                 Some { (Option.get jl) with reachable = true })
          |> analyse_blk f.entry
        in
        (f, blksdefs))
      program.functions
  in
  (* Maintenant qu'on a les définitions, on peut mettre à jour les instructions *)
  let rec update_ins (defs : def_tbl) ((ins, fins) : block) : block =
    match ins with
    (* On traite l'instruction i *)
    | i :: q ->
        (* On récupère les instructions modifiés *)
        let prev_ins, prev_fins = update_ins (analyse_ins defs i) (q, fins) in
        (* Si r est constant, on simplifie l'instruction en conséquence *)
        let set_r_ins (r : register) (v : value option) =
          match (get_val defs (Reg r), v) with
          | Some a, Some b when a = b -> prev_ins
          | None, _ | Some _, _ -> (
              match v with
              | None -> i :: prev_ins
              | Some (VInt a) -> Cst (r, a) :: prev_ins
              | Some (VLoc a) -> Addr (r, a) :: prev_ins
              | Some VNull -> Cst (r, 0) :: prev_ins)
        in
        (* Si on store une constante qui est déjà utilisé, pas besoin de la mettre à jour *)
        let store_r_ins (prev_v : value option) (v : value option) =
          match (prev_v, v) with
          | Some a, Some b when a = b -> prev_ins
          | _ -> i :: prev_ins
        in
        (* On essaye de simplifier une opération de Store à une adresse addr_b*)
        let store_addr (addr_v : value option) (vr : register) =
          match addr_v with
          (* Adresse invalide où adresse inconnue *)
          | None | Some (VInt _) | Some VNull -> i :: prev_ins
          (* Adresse connue *)
          | Some (VLoc l) as addr -> (
              match get_val_at_addr defs addr with
              | Some v when Some v = get_val defs (Reg vr) -> prev_ins
              | _ -> Store (l, vr) :: prev_ins)
        in
        (* On modifie (au besoin) l'instruction *)
        ( (match i with
          | Store (l, r) ->
              store_r_ins (get_val defs (Loc l)) (get_val defs (Reg r))
          | StoreI (l, r) -> store_addr (get_val defs (Loc l)) r
          | StoreR (rres, r) -> store_addr (get_val defs (Reg rres)) r
          | Load (r, l) -> get_val defs (Loc l) |> set_r_ins r
          | Monop (res, op, r) ->
              get_val defs (Reg r) |> analyse_monop defs op |> set_r_ins res
          | Binop (res, op, r1, r2) ->
              analyse_binnop defs op (get_val defs (Reg r1))
                (get_val defs (Reg r2))
              |> set_r_ins res
          | Cst (r, i) -> store_r_ins (get_val defs (Reg r)) (Some (VInt i))
          | Addr (r, l) -> store_r_ins (get_val defs (Reg r)) (Some (VLoc l))
          | _ -> i :: prev_ins),
          prev_fins )
    | [] -> (
        match fins with
        | Jmp _ | Return _ -> ([], fins)
        (* On traite les JmpC constants *)
        | JmpC (c, lt, lf) ->
            ( [],
              match get_val_bool defs (Reg c) with
              | Some true -> Jmp lt
              | Some false -> Jmp lf
              | None -> fins ))
  in
  (* Une fois que l'on a calculé les constantes des blocs, on peut les mettre à jour*)
  let functions =
    blocs
    |> List.map (fun ((f : function_cfg), (blkdefs : blkdef_tbl)) ->
           let topo = topo_sort f.entry f.graph in
           let graph =
             (* On remplace les blocs de la fonction f *)
             List.fold_left
               (fun (g : graph) (l : label) ->
                 let data = BlockMap.find l blkdefs in
                 replace_block l
                   (update_ins (Option.value data.init_def ~default:empty_defs))
                   g)
               f.graph topo
           in
           { f with graph })
  in
  { program with functions }
