open Cfg
open PrettyPrinter
open Analysis

module BlockMap = Map.Make (struct
  type t = label

  let compare = compare
end)

(*
 * Le premier bool représente si on a besoin de la variable ou du registre plus tard et
 * Le deuxième bool représente si c'est une variable qui a perdu son adresse
 *)
type value = bool * bool

(* Formate une valeur *)
let pp_value p =
  let open Format in
  function
  | None -> fprintf p "Not needed"
  | Some (a, b) -> fprintf p "(needed: %b, unsafe: %b)" a b

type req_tbl = value LocMap.t

type blk_data = {
  end_requirements : req_tbl;
  pos : int;
  next_topo_label : label option;
}

type blkreq_tbl = blk_data BlockMap.t

(* Marque la position pos comme demandé (ou pas suivant needed) dans la suite du bloc *)
let set_needed (pos : position) (needed : bool) (reqs : req_tbl) : req_tbl =
  (* Format.printf "Set needed: %a: %b\n" pp_position pos needed; *)
  LocMap.update pos
    (fun v ->
      let a, b = Option.value v ~default:(false, false) in
      Some (needed, b))
    reqs

let set_unsafe (pos : position) (reqs : req_tbl) : req_tbl =
  (* Format.printf "Set unsafe: %a\n" pp_position pos; *)
  LocMap.update pos
    (fun v ->
      let a, b = Option.value v ~default:(false, false) in
      Some (a, true))
    reqs

(* Renvoie si on a besoin (potentiellement) de la position pos plus tard *)
let is_needed (pos : position) (reqs : req_tbl) : bool =
  match LocMap.find_opt pos reqs with None -> false | Some (a, _) -> a

(* On initialise les variables requises pour chaque bloc *)
let rec init_blkreqs (l : label list) (init_reqs : req_tbl) (pos : int)
    (blkdefs : blkreq_tbl) =
  match l with
  | [] -> blkdefs
  | [ x ] ->
      BlockMap.add x
        { end_requirements = init_reqs; pos; next_topo_label = None }
        blkdefs
  | x :: y :: q ->
      BlockMap.add x
        { end_requirements = init_reqs; pos; next_topo_label = Some y }
        blkdefs
      |> init_blkreqs (y :: q) init_reqs (pos + 1)

(* On fusionne les variables requises et on renvoie si "a" a changé *)
let merge_reqs (a : blk_data) (b : req_tbl) : bool * blk_data =
  let changed = ref false in
  let union_needed x y =
    match (x, y) with
    | Some (a1, a2), Some (b1, b2) ->
        let res = (a1 || b1, a2 || b2) in
        changed := res <> (a1, a2);
        Some res
    | None, Some v ->
        changed := v <> (false, false);
        Some v
    | v, None -> v
  in
  let res = LocMap.merge (Fun.const union_needed) a.end_requirements b in
  (!changed, { a with end_requirements = res })

(* On filtre le programme *)
let analyse (program : program) =
  (* On vient de faire une opération inconnue ou on appelle une fonction, on a sûrement besoin des
     variables globales / qui ont perdu leurs adresses *)
  let set_unsafe_needed (req : req_tbl) =
    LocMap.map (fun (a, b) -> (a || b, b)) req
    |> Fun.flip
         (List.fold_left (fun req g -> set_needed (Loc (Global g)) true req))
         program.globals
  in
  (* Analyse d'une instruction *)
  let analyse_ins (i : non_final_instruction) (reqs : req_tbl) : req_tbl =
    let basic_set (dest : position) (from : position) (req : req_tbl) : req_tbl
        =
      if is_needed dest reqs then
        set_needed dest false reqs |> set_needed from true
      else reqs
    in
    let basic_binop_set (dest : position) (from1 : position) (from2 : position)
        (req : req_tbl) : req_tbl =
      if is_needed dest reqs then
        set_needed dest false reqs |> set_needed from1 true
        |> set_needed from2 true
      else reqs
    in
    match i with
    | Cst (r, _) -> set_needed (Reg r) false reqs
    | Store (l, r) -> basic_set (Loc l) (Reg r) reqs
    | StoreI (l, r) ->
        set_unsafe_needed reqs |> set_needed (Reg r) true
        |> set_needed (Loc l) true
    | StoreR (r1, r2) ->
        set_unsafe_needed reqs |> set_needed (Reg r1) true
        |> set_needed (Reg r2) true
    | Load (r, l) -> basic_set (Reg r) (Loc l) reqs
    | Addr (r, l) -> set_needed (Reg r) false reqs
    | PushCallArg (r, _) -> set_needed (Reg r) true reqs
    | PopCallRes r -> set_needed (Reg r) false reqs
    | Monop (r1, _, r2) -> basic_set (Reg r1) (Reg r2) reqs
    | Binop (r1, _, r2, r3) -> basic_binop_set (Reg r1) (Reg r2) (Reg r3) reqs
    | CallR r -> set_unsafe_needed reqs |> set_needed (Reg r) true
    | Call _ -> set_unsafe_needed reqs
    | Nop | PrintString _ | FreeRegs _ -> reqs
  in
  (* Analyse des instructions non finales *)
  let analyse_fins (fins : final_instruction) (reqs : req_tbl) =
    match fins with
    | Return r -> set_unsafe_needed reqs |> set_needed (Reg r) true
    | JmpC (r, l, l') when l <> l' -> set_needed (Reg r) true reqs
    | JmpC _ | Jmp _ -> reqs
  in
  (* On enlève les blocs vides en sautant les blocs vides *)
  let remove_empty_jmp (graph : graph) (reqs : req_tbl) :
      final_instruction -> graph * final_instruction =
    let get_l (graph : graph) (next_l : label) =
      let rec fold_free_blk (dest : label) ins fr =
        match ins with
        | FreeRegs r :: q -> fold_free_blk dest q (RegSet.union fr r)
        | [] ->
            (* On indique que l'on a traité le bloc *)
            ( replace_block next_l (fun b -> ([], Jmp dest)) graph
              |> Ast2cfg.propagate_free_regs (Jmp dest) fr,
              dest )
        | l ->
            let graph =
              if fr = RegSet.empty then graph
              else
                replace_block next_l
                  (fun _ -> (FreeRegs fr :: l, Jmp dest))
                  graph
            in
            (* Pas possible *) (graph, next_l)
      in
      match get_block next_l graph with
      | ins, Jmp l -> fold_free_blk l ins RegSet.empty
      | ins, JmpC (_, a, b) when a = b -> fold_free_blk a ins RegSet.empty
      | _ -> (graph, next_l)
    in
    function
    | Return r -> (graph, Return r)
    | Jmp l ->
        let g, l = get_l graph l in
        (g, Jmp l)
    | JmpC (r, l1, l2) ->
        (* Techniquement, on pourrait refaire une analyse parce qu'on a plus besoin de r…
           On ne le transforme pas en Jmp car, on veut se rappeler que Reg r n'est plus utile. *)
        let g, l1 = get_l graph l1 in
        let g, l2 = get_l g l2 in
        (g, JmpC (r, l1, l2))
  in
  (* Supprime les instructions inutiles *)
  let update_ins (reqs : req_tbl) (graph : graph) ((ins, fins) : block) :
      graph * block =
    (* On renvoie si on doit garder l'instruction ou pas *)
    let needed (reqs : req_tbl) = function
      | (Cst (r, _) | Load (r, _) | Addr (r, _)) as i ->
          if is_needed (Reg r) reqs then Some i else None
      | Monop (r1, o, r2) ->
          if is_needed (Reg r1) reqs then Some (Monop (r1, o, r2)) else None
      | Binop (r1, o, r2, r3) ->
          if is_needed (Reg r1) reqs then Some (Binop (r1, o, r2, r3)) else None
      | Store (l, r) ->
          if is_needed (Loc l) reqs then Some (Store (l, r)) else None
      | StoreI (l, r) -> Some (StoreI (l, r))
      | StoreR (a, r) -> Some (StoreR (a, r))
      | PushCallArg (r, n) -> Some (PushCallArg (r, n))
      | PopCallRes r ->
          if is_needed (Reg r) reqs then Some (PopCallRes r) else None
      | CallR a -> Some (CallR a)
      | (Call _ | PrintString _ | FreeRegs _) as i -> Some i
      | Nop -> None
    in
    let _, ins =
      List.fold_right
        (fun (i : non_final_instruction)
             ((reqs : req_tbl), (nins : non_final_instruction list)) ->
          ( analyse_ins i reqs,
            match needed reqs i with Some i -> i :: nins | None -> nins ))
        ins
        (analyse_fins fins reqs, [])
    in
    let graph, fins = remove_empty_jmp graph reqs fins in
    (* On itère sur les instructions et on garde seulement celles dont on a besoin *)
    (graph, (ins, fins))
  in
  (* On met à jour chaque bloque *)
  let blks =
    List.map
      (fun (f : function_cfg) ->
        (* On itère sur chaque fonction *)
        (* On veut traiter les blocs par le bas, on renverse le tri topologique *)
        let topo = List.rev (topo_sort f.entry f.graph) in
        (* On analyse le bloc l *)
        let rec analyse_blk (l : label) (blkreqs : blkreq_tbl) =
          (* Format.eprintf "\n# Analyse bloc: %a\n\n" pp_label l; *)
          let data = BlockMap.find l blkreqs in
          (* On met à jour les variables requises des blocs précédents *)
          let update_reqs (reqs : req_tbl) (prev_label : label)
              (next_label : (label * int) option ref) (blkreqs : blkreq_tbl) :
              blkreq_tbl =
            BlockMap.update prev_label
              (function
                | None ->
                    None
                    (* block non atteignable en descendant mais atteignable par le haut, on peut l'ignorer *)
                | Some prev_data ->
                    let changed, new_j_defs = merge_reqs prev_data reqs in
                    if changed && prev_data.pos <= data.pos then begin
                      next_label :=
                        match !next_label with
                        | None -> Some (prev_label, prev_data.pos)
                        | Some (_, p) when p > prev_data.pos ->
                            Some (prev_label, prev_data.pos)
                        | x -> x
                    end;
                    Some new_j_defs)
              blkreqs
          in
          let ins, fins = get_block l f.graph in
          let data = BlockMap.find l blkreqs in
          let reqs =
            data.end_requirements |> analyse_fins fins
            |> List.fold_right analyse_ins ins
          in
          let next_topo_label =
            ref
              (Option.bind data.next_topo_label (fun x ->
                   Some (x, data.pos + 1)))
          in
          let blkreqs =
            get_predecessors l f.graph
            |> List.fold_left
                 (fun blkreqs (pl, _) ->
                   update_reqs reqs pl next_topo_label blkreqs)
                 blkreqs
          in
          match !next_topo_label with
          | Some (nl, _) -> analyse_blk nl blkreqs
          | None -> (f, blkreqs)
        in
        (* On analyse chaque bloc de la fonction *)
        let init_req =
          List.fold_left
            (fun reqs l ->
              let ins, _ = get_block l f.graph in
              (* On récupère la liste des variables qui donne leur adresse *)
              List.fold_left
                (fun reqs -> function
                  | Addr (_, l) -> set_unsafe (Loc l) reqs | _ -> reqs)
                reqs ins)
            LocMap.empty topo
        in
        let blksreqs =
          init_blkreqs topo init_req 0 BlockMap.empty
          |> analyse_blk (List.hd topo)
        in
        blksreqs)
      program.functions
  in

  (* Maintenant que l'on a analysé les blocs, on peut les mettre à jour en supprimant les
     instructions inutiles *)
  let functions =
    List.map
      (fun ((f : function_cfg), (blkdefs : blkreq_tbl)) ->
        let topo = topo_sort f.entry f.graph |> List.rev in
        let graph =
          List.fold_left
            (fun (g : graph) (l : label) ->
              let data = BlockMap.find l blkdefs in
              let g, b = update_ins data.end_requirements g (get_block l g) in
              replace_block l (fun _ -> b) g)
            f.graph topo
        in
        { f with graph })
      blks
  in
  { program with functions }
