open Cfg
open PrettyPrinter
open Analysis

module BlockMap = Map.Make (struct
  type t = label

  let compare = compare
end)

type def_tbl = bool LocMap.t
type blkdef_tbl = def_tbl BlockMap.t

(* Renvoie a ou b (c'est à dire le premier qui n'est pas None)*)
let or_option (a : 'a option) (b : 'a option) : 'a option =
  match a with None -> b | Some _ -> a

(* On analyse les initialisation *)
let analyse (program : program) =
  (* Renvoie si la variable v a été initialisé *)
  let is_init (t : def_tbl) (v : position) : bool =
    LocMap.find_opt v t |> Option.value ~default:false
  in
  (* On itère sur chaque fonction *)
  List.iter
    (fun (f : function_cfg) ->
      let check_def_blk (blocks_defs : blkdef_tbl) (l : label) : blkdef_tbl =
        (* On itère sur chaque bloc *)
        (* On marque la position p comme défini *)
        let define (p : position) (deftbl : def_tbl) : def_tbl =
          (* Format.eprintf "Defined: %a\n" pp_position p; *)
          LocMap.add p true deftbl
        in
        (* On utilise la position p, on s'assure que p est défini *)
        let assert_def (p : position) (deftbl : def_tbl) : def_tbl =
          (* Format.eprintf "Used: %a\n" pp_position p; *)
          if is_init deftbl p then deftbl
          else (
            Format.eprintf "warning: init: %a may not be initialised in %a\n"
              pp_position p pp_label l;
            (* Pour éviter de remettre l'avertissement, on la défini artificiellement *)
            define p deftbl)
        in

        (* On analyse une instruction *)
        let add_ins_in_def_tbl (deftbl : def_tbl) (i : non_final_instruction) :
            def_tbl =
          match i with
          | Store ((Local _ as l), r) ->
              assert_def (Reg r) deftbl |> define (Loc l)
          | StoreI ((Local _ as l), r) ->
              assert_def (Reg r) deftbl |> assert_def (Loc l)
          | StoreR (r1, r2) -> assert_def (Reg r1) deftbl |> assert_def (Reg r2)
          | Load (r, (Local _ as l)) ->
              assert_def (Loc l) deftbl |> define (Reg r)
          | Cst (r, _)
          | Load (r, _)
          | Addr (r, _)
          | PopCallRes r ->
              define (Reg r) deftbl
          | Store (_, r) | StoreI (_, r) | PushCallArg (r, _) | CallR r ->
              assert_def (Reg r) deftbl
          | Monop (r1, _, r2) -> assert_def (Reg r2) deftbl |> define (Reg r1)
          | Binop (r1, _, r2, r3) ->
              assert_def (Reg r2) deftbl |> assert_def (Reg r3)
              |> define (Reg r1)
          | Nop | Call _ | PrintString _ | FreeRegs _ -> deftbl
        in
        (* On analyse une instruction finale *)
        let assert_fins_in_def_tbl (i : final_instruction) (deftbl : def_tbl) =
          match i with
          | Return r -> assert_def (Reg r) deftbl
          | JmpC (r, _, _) -> assert_def (Reg r) deftbl
          | Jmp _ -> deftbl
        in
        let ins, fins = get_block l f.graph in
        let prev_blks = get_predecessors l f.graph in
        (* On a terminé de vérifier le bloc, on passe les informations de définitions au blocs
           suivants *)
        let start_defs =
          (* Seulement les boucles ne seront pas dans blkdef, et comme on ne peut pas déinitialiser
             une variable, on peut l'ignorer car initialisé ssi initialisé dans le premier
             tour de boucle *)
          List.filter_map
            (fun (l, _) -> BlockMap.find_opt l blocks_defs)
            prev_blks
          |> function
          | [] -> LocMap.empty
          | x :: q ->
              List.fold_left
                (LocMap.merge (fun _ a b ->
                     match (a, b) with
                     | Some true, Some true -> Some true
                     | _ -> None))
                x q
        in
        let deftbl =
          List.fold_left add_ins_in_def_tbl start_defs ins
          |> assert_fins_in_def_tbl fins
        in
        BlockMap.add l deftbl blocks_defs
      in
      let t = topo_sort f.entry f.graph in
      ignore @@ List.fold_left check_def_blk BlockMap.empty t)
    program.functions
