open Cast
open Cfg

exception Anomaly of string

type next_instr = Label of label | Block of block

type data = {
  globals : string list; (* les variables globales sans les fonctions *)
  arg_map : int VarMap.t;
  (* liste des variables locales, indice dans la pile et identifiant unique*)
  local_map : (int * int) VarMap.t;
  nb_args : int; (* nombre d'arguments de la fonction en cours *)
  nb_local_in_use : int; (* nombre de variables locales dans local_map *)
  next : next_instr; (* instruction suivante à exécuter *)
  graph : graph; (* graph de la fonction en cours de construction *)
}

(* On crée un nouveau registre *)
let new_reg : unit -> register =
  let x = ref 0 in
  fun () ->
    incr x;
    reg_of_string (Printf.sprintf "reg_%d" !x)

(* On veut exécuter nfi après next de data, on renvoie un data modifié en conséquence *)
let add_data_ins (nfi : non_final_instruction) (data : data) : data =
  {
    data with
    next =
      (match data.next with
      | Label l -> Block ([ nfi ], Jmp l)
      | Block (ins, fi) -> Block (nfi :: ins, fi));
  }

let free_regs (l : register list) : data -> data =
  add_data_ins (FreeRegs (RegSet.of_list l))

(* On ferme le bloc courant si on en a un et on pointe vers son label *)
let force_next_label (data : data) : data =
  match data.next with
  | Label _ -> data
  | Block b ->
      let l, graph = add_block b data.graph in
      { data with graph; next = Label l }

(* On propage les FreeRegs *)
let propagate_free_regs (fi : final_instruction) (fr : RegSet.t) (graph : graph)
    : graph =
  if fr = RegSet.empty then graph
  else
    let aux l graph =
      replace_block l
        (function
          | FreeRegs fr' :: q, fi -> (FreeRegs (RegSet.union fr fr') :: q, fi)
          | l, fi -> (FreeRegs fr :: l, fi))
        graph
    in
    match fi with
    | Jmp l -> aux l graph
    | JmpC (r, l, l') -> aux l graph |> aux l'
    | Return _ -> graph

(* On veut un bloc vide, on crée un label au besoin (mais on permet de réutiliser
   une instruction finale seule). On ne veut pas répéter un Ret car le dépilement du cadre est assez coûteux *)
let prefers_next_label (data : data) : data =
  let rec fold_free_regs fi b fr =
    match b with
    | [] -> { data with graph = propagate_free_regs fi fr data.graph }
    | FreeRegs f :: q -> fold_free_regs fi q (RegSet.union fr f)
    | _ ->
        let l, graph =
          add_block
            ((if fr = RegSet.empty then b else FreeRegs fr :: b), fi)
            data.graph
        in
        { data with graph; next = Label l }
  in
  match data.next with
  | Label _ -> data
  | Block (li, (Jmp _ as fi)) | Block (li, (JmpC _ as fi)) ->
      fold_free_regs fi li RegSet.empty
  | Block b ->
      let l, graph = add_block b data.graph in
      { data with graph; next = Label l }

(* On termine le bloc en cours et on l'ajoute (cette fonction est utilisé pour des blocs
   temporaires alors que force_next_label est utilisé pour des blocs que l'on réutilise cf CIF) *)
let get_next_label (data : data) : label * graph =
  match data.next with
  | Label l -> (l, data.graph)
  | Block b -> add_block b data.graph

(* Renvoie le nom de variable associé à une expression (par exemple utile pour traiter &v) *)
let get_var_expr (e : expr) =
  match e with
  | VAR n -> n
  | _ -> raise (Anomaly "Une variable était attendue.")

(* Renvoie la location d'une variable à partir de son nom*)
let get_var_location (n : string) (data : data) : location =
  (* Fonction qui renvoie Some (fa a) si a est pas None, sinon si b n'est pas None, Some (fb b)
     sinon None *)
  let option_or_bind a fa b fb =
    match a with Some _ -> Option.bind a fa | None -> Option.bind b fb
  in
  (* On cherche dans les variables locales puis les arguments puis les variables globales et sinon
     c'est un nom de fonction car le programme est sensé être bien typé *)
  match
    option_or_bind
      (VarMap.find_opt n data.local_map)
      (fun (i, j) -> Some (Local (i, j)))
      (VarMap.find_opt n data.arg_map)
      (fun i -> Some (Param i))
  with
  | Some l -> l
  | None when List.mem n data.globals -> Global n
  | None -> FuncLabel n (* car le programme est bien typé *)

(* On traite une déclaration globale en CFG *)
let rec transform_global_decl ({ globals; functions } : program) :
    var_declaration -> program = function
  | CDECL (_, name, _) -> { globals = name :: globals; functions }
  | CFUN (_, name, args, _, (_, code)) ->
      let function_cfg = transform_fun_decl globals name args code in
      { globals; functions = function_cfg :: functions }

(* On transforme une fonction en CFG *)
and transform_fun_decl (globals : string list) (name : string)
    (args : var_declaration list) (code : code) : function_cfg =
  let _, arg_map =
    List.fold_left
      (fun (i, m) decl ->
        match decl with
        | CDECL (_, x, _) -> (i + 1, VarMap.add x i m)
        | _ ->
            raise
              (Anomaly
                 "Déclaration de fonction en tant qu'argument de fonction \
                  impossible !"))
      (0, VarMap.empty) args
  in
  (* Comme tous les programme return correctement ça ne devrait JAMAIS arriver de conserver ce return *)
  let ret_reg = reg_of_string "NO RETURN !" in
  let block = ([], Return ret_reg) in
  let next = Block block in
  let local_map = VarMap.empty in
  let nb_args = List.length args in
  let data =
    transform_code code
      {
        globals;
        arg_map;
        local_map;
        next;
        nb_args;
        nb_local_in_use = 0;
        graph = empty_graph;
      }
  in
  match data.next with
  | Label l -> { name; graph = data.graph; entry = l; nb_args }
  | Block b ->
      let entry, graph = add_block b data.graph in
      { name; graph; entry; nb_args }

(* On traite la déclaration d'une nouvelle variable en lui allouant une place sur la pile *)
and convert_local_decl : data -> var_declaration -> data =
  let counter = ref (-1) in
  fun data -> function
    | CDECL (_, name, _) ->
        incr counter;
        {
          data with
          nb_local_in_use = data.nb_local_in_use + 1;
          local_map =
            VarMap.add name (data.nb_local_in_use, !counter) data.local_map;
        }
    | _ ->
        raise
          (Anomaly
             "Déclaration d'une fonction en tant que variable locale \
              impossible !")

(* On continue de transformer le code c  *)
and transform_code (c : code) (data : data) : data =
  match c with
  | CBLOCK (var_decls, loc_code) ->
      let outer_locals = data.local_map in
      let outer_locals_nb = data.nb_local_in_use in
      let code = List.map snd loc_code in
      let data =
        (* On déclare les variables locales *)
        List.fold_left convert_local_decl data var_decls
        (* on exécute les commandes du blocs puis on  *)
        |> List.fold_right transform_code code
      in
      (*il faut que les variables locales définies dans le block ne soient plus retenues,
        il faudra également dépiler celles ci*)
      { data with local_map = outer_locals; nb_local_in_use = outer_locals_nb }
  | CEXPR (_, e) ->
      let r = new_reg () in
      free_regs [ r ] data |> transform_expr r e
  | CIF ((_, c), (_, t), (_, f)) ->
      (* Pour éviter de recopier le code dans chaque branche, on demande de fermer les blocs trop long *)
      let data = prefers_next_label data in
      (* On traite chaque branche *)
      let true_label, graph = transform_code t data |> get_next_label in
      let false_label, graph =
        transform_code f { data with graph } |> get_next_label
      in
      let r = new_reg () in
      { data with graph; next = Block ([], JmpC (r, true_label, false_label)) }
      |> transform_expr r c
  | CWHILE ((_, c), (_, b)) ->
      (* On génère un label pour continuer l'exécution après *)
      let data = force_next_label data in
      let end_label = fst (get_next_label data) in
      let r = new_reg () in
      let l, graph =
        loop (fun l ->
            { data with next = Block ([], JmpC (r, l, end_label)) }
            |> transform_expr r c |> transform_code b |> get_next_label)
      in
      { data with graph; next = Block ([], JmpC (r, l, end_label)) }
      |> transform_expr r c
  | CRETURN ret_v -> (
      match ret_v with
      | None ->
          let r = new_reg () in
          { data with next = Block ([ Cst (r, 0) ], Return r) }
      | Some (_, e) ->
          let r = new_reg () in
          { data with next = Block ([], Return r) } |> transform_expr r e)
(* | _ -> data *)

and transform_expr (r : register) (e : expr) (data : data) : data =
  match e with
  | NULL -> add_data_ins (Cst (r, 0)) data
  | CST i -> add_data_ins (Cst (r, i)) data
  | STRING s -> add_data_ins (Load (r, StringCst s)) data
  | VAR v -> add_data_ins (Load (r, get_var_location v data)) data
  | SET_VAR (v, (_, e)) ->
      data
      |> add_data_ins (Store (get_var_location v data, r))
      |> transform_expr r e
  | SET_VAL (v, (_, e)) ->
      data
      |> add_data_ins (StoreI (get_var_location v data, r))
      |> transform_expr r e
  | CALL ("print_string", [ (_, STRING s) ]) ->
      data |> add_data_ins (PrintString s)
  | CALL (fn, el) -> begin
      (* On suppose pour l'instant qu'une fonction retourne la valeur de résultat sur la pile et
         qu'elle s'occupe d'enlever ses arguments *)
      let data =
        match get_var_location fn data with
        | FuncLabel fl ->
            data |> add_data_ins (PopCallRes r) |> add_data_ins (Call fl)
        | l ->
            let nr = new_reg () in
            data |> free_regs [ nr ]
            |> add_data_ins (PopCallRes r)
            |> add_data_ins (CallR nr)
            |> add_data_ins (Load (nr, l))
      in
      let (data, _), args_reg =
        List.fold_left_map
          (fun (d, n) _ ->
            let r = new_reg () in
            ( (d |> free_regs [ r ] |> add_data_ins (PushCallArg (r, n)), n + 1),
              r ))
          (data, 0) el
      in
      List.fold_left2
        (fun d (_, e) r -> d |> transform_expr r e)
        data el args_reg
    end
  | OP1 (op, (_, e)) -> transform_op1 r data op e
  | OP2 (op, (_, e1), (_, e2)) ->
      let nr = new_reg () in
      data |> free_regs [ nr ]
      |> add_data_ins
           (Binop
              ( r,
                (match op with
                | S_MUL -> MUL
                | S_DIV -> DIV
                | S_MOD -> MOD
                | S_ADD -> ADD
                | S_SUB -> SUB),
                r,
                nr ))
      |> transform_expr nr e2 |> transform_expr r e1
  | CMP (op, (_, e1), (_, e2)) ->
      let nr = new_reg () in
      data |> free_regs [ nr ]
      |> add_data_ins
           (Binop
              ( r,
                (match op with
                | C_EQ -> CMP_EQ
                | C_LT -> CMP_LT
                | C_LE -> CMP_LE),
                r,
                nr ))
      |> transform_expr nr e2 |> transform_expr r e1
  | EIF ((_, c), (_, t), (_, f)) ->
      (* Même chose que pour CIF, on traite chaque branche séparément… *)
      let data = prefers_next_label data in
      let true_label, graph = data |> transform_expr r t |> get_next_label in
      let false_label, graph =
        { data with graph } |> transform_expr r f |> get_next_label
      in
      { data with graph; next = Block ([], JmpC (r, true_label, false_label)) }
      |> transform_expr r c
  | ESEQ s -> List.fold_right (fun (_, e) -> transform_expr r e) s data

(* On convertit une opération unaire *)
and transform_op1 (r : register) (data : data) (op : mon_op) (e : expr) : data =
  match op with
  | M_MINUS -> transform_expr r e (add_data_ins (Monop (r, MINUS, r)) data)
  | M_NOT -> transform_expr r e (add_data_ins (Monop (r, NOT, r)) data)
  | M_POST_INC ->
      let nr = new_reg () in
      let data = free_regs [ nr ] data in
      transform_pre_inc_dec r nr data 1 e
  | M_POST_DEC ->
      let nr = new_reg () in
      let data = free_regs [ nr ] data in
      transform_pre_inc_dec r nr data (-1) e
  | M_PRE_INC -> transform_pre_inc_dec r r data 1 e
  | M_PRE_DEC -> transform_pre_inc_dec r r data (-1) e
  | M_DEREF -> transform_expr r e (add_data_ins (Monop (r, DEREF, r)) data)
  | M_ADDR ->
      add_data_ins (Addr (r, get_var_location (get_var_expr e) data)) data

(* On s'occupe d'une opération unaire d'incrémentation ou de décrémentation *)
and transform_pre_inc_dec (prev : register) (post : register) (data : data)
    (off : int) (e : expr) : data =
  match e with
  (* v++ *)
  | VAR n ->
      let loc = get_var_location n data in
      add_data_ins (Store (loc, post)) data
      |> add_data_ins (Monop (post, ADDI off, prev))
      |> add_data_ins (Load (prev, loc))
      (* (DEREF truct)++ *)
  | OP1 (M_DEREF, (_, de)) ->
      let ptr = new_reg () in
      let data =
        free_regs [ ptr ] data
        |> add_data_ins (StoreR (ptr, post))
        |> add_data_ins (Monop (post, ADDI off, prev))
        |> add_data_ins (Monop (prev, DEREF, ptr))
      in
      transform_expr ptr de data
  | _ -> raise (Anomaly "Transformation ++ incorrecte.")

let transform_program : var_declaration list -> program =
  List.fold_left transform_global_decl { globals = []; functions = [] }
