type label = { label : string }

let label_of_func_name name = { label = name }

type register = { reg_name : string }

let reg_of_string reg_name = { reg_name }

type monop = ADDI of int | MOV | DEREF | MINUS | NOT
type binop = MUL | DIV | MOD | ADD | SUB | CMP_EQ | CMP_LT | CMP_LE

type location =
  | StringCst of string
  | FuncLabel of string
  | Global of string
  | Param of int
  | Local of int * int

type final_instruction =
  | Return of register
  | Jmp of label
  | JmpC of register * label * label

module RegSet = Set.Make (struct
  type t = register

  let compare a b = String.compare a.reg_name b.reg_name
end)

type non_final_instruction =
  | Nop
  | Cst of register * int
  | Store of location * register
  | StoreI of location * register
  | StoreR of register * register
  | Load of register * location
  | Addr of register * location
  | PushCallArg of register * int
  | PopCallRes of register
  | Monop of register * monop * register
  | Binop of register * binop * register * register
  | Call of string
  | CallR of register
  | PrintString of string
  | FreeRegs of RegSet.t

type block = non_final_instruction list * final_instruction

module VarMap = Map.Make (String)

module G = Map.Make (struct
  type t = label

  let compare = compare
end)

type graph = block G.t

let empty_graph : graph = G.empty

let new_label =
  let x = ref (-1) in
  fun () ->
    incr x;
    { label = Format.asprintf "L%i" !x }

let add_block b g =
  let l = new_label () in
  (l, G.add l b g)

let loop f =
  let l = new_label () in
  let l', g = f l in
  (l, G.add l (G.find l' g) (G.remove l' g))

let get_block l g =
  try G.find l g
  with _ ->
    failwith (Format.asprintf "Label %s not associated to any block" l.label)

let get_predecessors : label -> graph -> (label * block) list =
 fun l g ->
  G.filter
    (fun _ (_, final) ->
      match final with
      | JmpC (_, _, l') when l = l' -> true
      | (Jmp l' | JmpC (_, l', _)) when l = l' -> true
      | _ -> false)
    g
  |> G.bindings

let replace_block l f g =
  match G.find_opt l g with
  | Some b -> G.add l (f b) g
  | None -> failwith "Label not associated to any block"

(* Un tri topologique des labels atteignables depuis entry *)
let topo_sort (entry : label) (graph : graph) : label list =
  let visited = Hashtbl.create 10 in
  let rec visit (l : label) (acc : label list) : label list =
    Hashtbl.add visited l true;
    match snd (get_block l graph) with
    | Jmp l' -> l :: (if Hashtbl.mem visited l' then acc else visit l' acc)
    | JmpC (_, l1, l2) ->
        let acc = if Hashtbl.mem visited l1 then acc else visit l1 acc in
        let acc = if Hashtbl.mem visited l2 then acc else visit l2 acc in
        l :: acc
    | _ -> l :: acc
  in
  visit entry []

type function_cfg = {
  name : string;
  graph : graph;
  entry : label;
  nb_args : int;
}

type program = { globals : string list; functions : function_cfg list }

module PrettyPrinter = struct
  module S = Set.Make (struct
    type t = label

    let compare = compare
  end)

  let pp_register p { reg_name } =
    Format.fprintf p "%s" reg_name

  let pp_label p { label } = Format.fprintf p "%s" label

  let pp_monop p = function
    | ADDI i -> Format.fprintf p "%i +" i
    | MOV -> Format.fprintf p ""
    | MINUS -> Format.fprintf p "-"
    | NOT -> Format.fprintf p "!"
    | DEREF -> Format.fprintf p "*"

  let pp_binop p op =
    Format.fprintf p "%s"
      (match op with
      | MUL -> "*"
      | DIV -> "/"
      | MOD -> "%"
      | ADD -> "+"
      | SUB -> "-"
      | CMP_EQ -> "=="
      | CMP_LT -> "<"
      | CMP_LE -> "<=")

  let pp_location p =
    let open Format in
    function
    | StringCst s -> fprintf p "String %s" s
    | FuncLabel s -> fprintf p "FuncLabel %s" s
    | Global s -> fprintf p "Global %s" s
    | Param i -> fprintf p "Param %i" i
    | Local (i, j) -> fprintf p "Local %i(%i)" i j

  let pp_instr p =
    let open Format in
    function
    | Nop -> fprintf p "Nop"
    | Cst (r, i) -> fprintf p "%a <- %i" pp_register r i
    | Store (loc, r) -> fprintf p "%a <- %a" pp_location loc pp_register r
    | StoreI (loc, r) -> fprintf p "*%a <- %a" pp_location loc pp_register r
    | StoreR (r', r) -> fprintf p "*%a <- %a" pp_register r' pp_register r
    | Load (r, loc) -> fprintf p "%a <- %a" pp_register r pp_location loc
    | Addr (r, loc) -> fprintf p "%a <- &(%a)" pp_register r pp_location loc
    | PushCallArg (r, n) -> fprintf p "PushCallArg %d <- %a" n pp_register r
    | PopCallRes r -> fprintf p "PopCallRes %a" pp_register r
    | Monop (r1, monop, r2) ->
        fprintf p "%a <- %a %a" pp_register r1 pp_monop monop pp_register r2
    | Binop (r1, binop, r2, r3) ->
        fprintf p "%a <- %a %a %a" pp_register r1 pp_register r2 pp_binop binop
          pp_register r3
    | Call f -> fprintf p "Call %s" f
    | CallR r -> fprintf p "CallR %a" pp_register r
    | PrintString s -> fprintf p "PrintString %s" s
    | FreeRegs r ->
        fprintf p "FreeRegs [@[<hov>%a@]]"
          (pp_print_list ~pp_sep:(fun p _ -> fprintf p ",@;") pp_register)
          (RegSet.to_list r)

  let pp_instr_list = Format.pp_print_list pp_instr
  let s_seen = ref S.empty

  let rec pp_final p (label, final, graph) =
    match final with
    | Jmp l ->
        Format.fprintf p "Jmp\" shape=box]@;%a -> %a@;%a" pp_label label
          pp_label l pp_cfg (l, graph)
    | JmpC (r, l1, l2) ->
        Format.fprintf p "JmpC %a\" shape=box]@;%a -> %a@;%a -> %a@;%a@;%a"
          pp_register r pp_label label pp_label l1 pp_label label pp_label l2
          pp_cfg (l1, graph) pp_cfg (l2, graph)
    | Return r -> Format.fprintf p "Ret %a\" shape=box]" pp_register r

  and pp_cfg p (entry, graph) =
    if S.mem entry !s_seen then Format.fprintf p ""
    else
      let instr_list, final = get_block entry graph in
      (* Format.printf "tutu\n"; *)
      s_seen := S.add entry !s_seen;
      Format.fprintf p "%a [label=\"%a:\n%a\\n%a" pp_label entry pp_label entry
        pp_instr_list instr_list pp_final (entry, final, graph)

  let pp_func_cfg p { name; graph; entry } =
    s_seen := S.empty;
    Format.fprintf p "%s [label=\"FUN %s\" shape=box]@;%s -> %a@;%a" name name
      name pp_label entry pp_cfg (entry, graph);
    s_seen := S.empty

  let pp_instr_list_simple =
    Format.pp_print_list ~pp_sep:(fun p () -> Format.fprintf p "@;") pp_instr

  let rec pp_final_simple p (label, final, graph) =
    match final with
    | Jmp l ->
        Format.fprintf p "Jmp@;%a -> %a@;%a" pp_label label pp_label l
          pp_cfg_simple (l, graph)
    | JmpC (r, l1, l2) ->
        Format.fprintf p "JmpC %a@;%a -> %a@;%a -> %a@;%a@;%a" pp_register r
          pp_label label pp_label l1 pp_label label pp_label l2 pp_cfg_simple
          (l1, graph) pp_cfg_simple (l2, graph)
    | Return r -> Format.fprintf p "Ret %a" pp_register r

  and pp_cfg_simple p (entry, graph) =
    if S.mem entry !s_seen then Format.fprintf p ""
    else
      let instr_list, final = get_block entry graph in
      (* Format.printf "tutu\n"; *)
      s_seen := S.add entry !s_seen;
      Format.fprintf p "- %a:@;%a@;%a" pp_label entry pp_instr_list_simple
        instr_list pp_final_simple (entry, final, graph)

  let pp_func_cfg_simple p { name; graph; entry } =
    s_seen := S.empty;
    Format.fprintf p "@[<v2>FUN %s {@;%s -> %a;@;%a@]@;}" name name pp_label
      entry pp_cfg_simple (entry, graph);
    s_seen := S.empty
end

let dump_dot filename { functions; _ } =
  let oc = open_out filename in
  let fmt = Format.formatter_of_out_channel oc in
  Format.fprintf fmt "@[<hv 2>digraph G {@;%a@]\n}@]"
    (Format.pp_print_list
       ~pp_sep:(fun p () -> Format.fprintf p "@;")
       PrettyPrinter.pp_func_cfg)
    functions

let dump_dot_simple filename { functions; _ } =
  let oc = open_out filename in
  let fmt = Format.formatter_of_out_channel oc in
  Format.fprintf fmt "%a@."
    (Format.pp_print_list
       ~pp_sep:(fun p () -> Format.fprintf p "\n")
       PrettyPrinter.pp_func_cfg_simple)
    functions
