open Cfg

exception Error of string

type asm_glb_cst_id = int (* Identifiant de constante *)
type asm_param_local_id = int (* Identifiant de paramètre / variable locale *)
type asm_cfg_reg = int (* Offset depuis le haut de la pile *)
type asm_real_reg = int (* Registre reel R0 à R7 *)

type glbcst =
  | IntCst of int
  | StringCst of string
  | Global of string
  | FuncLabel of label

module GlbCstMap = Map.Make (struct
  type t = glbcst

  let compare = compare
end)

module CfgRegMap = Map.Make (String)
module FunNameMap = Map.Make (String)

type asm_location =
  | GlobalCst of asm_glb_cst_id
  | Reg of asm_cfg_reg
  | ParamLocal of asm_param_local_id

type asm_ins =
  | Nop
  | LoadSmallCst of asm_cfg_reg * int
  | Load of asm_cfg_reg * asm_location
  | Store of asm_location * asm_cfg_reg
  | StoreI of asm_location * asm_cfg_reg
  | StoreR of asm_cfg_reg * asm_cfg_reg
  | Monop of asm_cfg_reg * monop * asm_cfg_reg
  | Binop of asm_cfg_reg * binop * asm_cfg_reg * asm_cfg_reg
  | Addr of asm_cfg_reg * asm_location
  | PushArg of asm_cfg_reg * asm_param_local_id
  | PopRes of asm_cfg_reg
  | Call of label
  | CallR of asm_cfg_reg
  | PushStack of
      int * int (* nb d'arguments+variables locales et nb de registres *)
  | Ret of asm_cfg_reg
  | Jmp of label
  | JmpC of asm_cfg_reg * label * label
  | PrintString of asm_glb_cst_id

type asm_block = label * asm_ins list
type asm_function = { blocks : asm_block list; name : string }

(* Données propagées lors de la première passe *)
type data = {
  glb_cst_table : int GlbCstMap.t;
  nb_glb_cst : int;
  fun_nb_args : int;
  fun_nb_locals_max : int;
  reg_table : int CfgRegMap.t;
  reg_free : int list;
  nb_reg_used : int;
  funs_entry_label : label FunNameMap.t;
}

let nb_regs = 8
let reg_fun_ret_addr : asm_real_reg = 7
let reg_stack : asm_real_reg = 6
let reg_glbcst : asm_real_reg = 5
let reg_fun_stack : asm_real_reg = 4
let reg_str_tmp : asm_real_reg = 3
let reg_fun_ret_val : asm_real_reg = 0
let start_stack = 0xFE00 - 1
let pp_glb_cst_id p (c : asm_glb_cst_id) = Format.fprintf p "GC%d" c
let pp_param_local_id p (pl : asm_param_local_id) = Format.fprintf p "PL%d" pl
let pp_cfg_reg p (r : asm_cfg_reg) = Format.fprintf p "reg_%d" r
let pp_real_reg p (r : asm_real_reg) = Format.fprintf p "R%d" r

let pp_location p (l : asm_location) =
  match l with
  | GlobalCst i -> pp_glb_cst_id p i
  | Reg r -> pp_cfg_reg p r
  | ParamLocal pl -> pp_param_local_id p pl

(* Print de débug pour les asm_ins *)
let pp_asm_ins p bi =
  let open Format in
  match bi with
  | Nop -> fprintf p "Nop"
  | LoadSmallCst (r, n) -> fprintf p "Load %a %d" pp_cfg_reg r n
  | Load (r, l) -> fprintf p "Load %a %a" pp_cfg_reg r pp_location l
  | Store (l, r) | StoreI (l, r) ->
      fprintf p "Store %a %a" pp_location l pp_cfg_reg r
  | StoreR (r, r') -> fprintf p "Store %a %a" pp_cfg_reg r pp_cfg_reg r'
  | Addr (r, l) -> fprintf p "Addr %a %a" pp_cfg_reg r pp_location l
  | Monop (r1, o, r2) ->
      fprintf p "Monop %a %a %a" pp_cfg_reg r1 PrettyPrinter.pp_monop o
        pp_cfg_reg r2
  | Binop (r1, o, r2, r3) ->
      fprintf p "Binop %a %a %a %a" pp_cfg_reg r1 PrettyPrinter.pp_binop o
        pp_cfg_reg r2 pp_cfg_reg r3
  | PushArg (r, i) -> fprintf p "PushArg %a %a" pp_param_local_id i pp_cfg_reg r
  | PushStack (i, j) -> fprintf p "PushStack PL:%d Regs:%d" i j
  | Call l -> fprintf p "Call %a" PrettyPrinter.pp_label l
  | CallR r -> fprintf p "CallR %a" pp_cfg_reg r
  | Ret r -> fprintf p "Ret %a" pp_cfg_reg r
  | PopRes r -> fprintf p "PopRes %a" pp_cfg_reg r
  | Jmp l -> fprintf p "Jmp %a" PrettyPrinter.pp_label l
  | JmpC (r, l1, l2) ->
      let open PrettyPrinter in
      fprintf p "JmpC %a %a %a" pp_cfg_reg r pp_label l1 pp_label l2
  | PrintString i -> fprintf p "PrintString %a" pp_glb_cst_id i

let pp_asm_block p (l, ins) =
  Format.(
    fprintf p "%a: @[<v>@;%a@]@;" PrettyPrinter.pp_label l
      (pp_print_list pp_asm_ins) ins)

let pp_asm_function p f =
  Format.(
    fprintf p "%s @;@[<v>%a@]@;" f.name (pp_print_list pp_asm_block) f.blocks)

(* On synthétise un LDR qui est potentiellement long *)
let rec pp_synth_ldr p (r, r', off) =
  let open Format in
  if -32 <= off && off < 32 then
    fprintf p "LDR %a %a #%d" pp_real_reg r pp_real_reg r' off
  else
    let doff = if off > 0 then 15 else -16 in
    fprintf p "ADD %a %a #%d@;" pp_real_reg r pp_real_reg r' doff;
    pp_synth_ldr p (r, r, off - doff)

(* On synthétise un STR qui est potentiellement long *)

let rec pp_synth_str p (r, r', off) =
  let open Format in
  if -32 <= off && off < 32 then
    fprintf p "STR %a %a #%d" pp_real_reg r pp_real_reg r' off
  else
    let doff = if off > 0 then 15 else -16 in
    fprintf p "ADD %a %a #%d@;" pp_real_reg reg_str_tmp pp_real_reg r' doff;
    pp_synth_str p (r, reg_str_tmp, off - doff)

(* On charge une location dans rr *)

let pp_synth_load_loc p ((rr : asm_real_reg), (l : asm_location)) =
  match l with
  | ParamLocal i -> pp_synth_ldr p (rr, reg_fun_stack, -i)
  | Reg r -> pp_synth_ldr p (rr, reg_stack, r)
  | GlobalCst g -> pp_synth_ldr p (rr, reg_glbcst, g)

(* On sauvegarde une location dans rr *)

let pp_synth_store_loc p ((l : asm_location), (rr : asm_real_reg)) =
  match l with
  | ParamLocal i -> pp_synth_str p (rr, reg_fun_stack, -i)
  | Reg r -> pp_synth_str p (rr, reg_stack, r)
  | GlobalCst g -> pp_synth_str p (rr, reg_glbcst, g)

let pp_synth_jsr p (label : label) =
  let open Format in
  if !Usage.long_jmp then
    fprintf p "LD R3 #1@;JSRR R3@;.FILL %a" PrettyPrinter.pp_label label
  else fprintf p "JSR %a" PrettyPrinter.pp_label label

(* Synthétisation d'un grand add *)
let rec pp_big_reg_add p (r, r', n) =
  if -16 <= n && n < 16 then
    Format.fprintf p "ADD %a %a #%d" pp_real_reg r pp_real_reg r' n
  else
    let diff = if n < 0 then -16 else 15 in
    Format.fprintf p "ADD %a %a #%d" pp_real_reg r pp_real_reg r' diff;
    if n - diff <> 0 then
      Format.fprintf p "@;%a" pp_big_reg_add (r, r', n - diff)

(* On charge l'adresse d'une location dans r *)
let pp_synth_load_addr p (r, l) =
  match l with
  | ParamLocal n ->
      Format.fprintf p "AND %a %a #0@;" pp_real_reg r pp_real_reg r;
      pp_big_reg_add p (r, reg_fun_stack, -n)
  | Reg n ->
      Format.fprintf p "AND %a %a #0@;" pp_real_reg r pp_real_reg r;
      pp_big_reg_add p (r, reg_stack, n)
  | GlobalCst g ->
      Format.fprintf p "AND %a %a #0@;" pp_real_reg r pp_real_reg r;
      pp_big_reg_add p (r, reg_glbcst, g)

(* Fonction de multiplication *)
let pp_synth_mult p () =
  let open Format in
  (* R2 <- R0 * R1 *)
  fprintf p "MULT: @;@[<v4>    ";
  if !Usage.long_jmp then
    fprintf p "ADD %a %a #1 @;" pp_real_reg reg_fun_ret_addr pp_real_reg
      reg_fun_ret_addr;
  fprintf p "AND R2 R2 #0@;";
  fprintf p "ADD R1 R1 #0@;";
  fprintf p "BRp MULT_LOOP@;";
  fprintf p "BRz MULT_END@;";
  fprintf p "NOT R1 R1@;";
  fprintf p "ADD R1 R1 #1@;";
  fprintf p "NOT R0 R0@;";
  fprintf p "ADD R0 R0 #1@;";
  fprintf p "@]@;";

  fprintf p "MULT_LOOP: @;@[<v4>    ";
  fprintf p "ADD R2 R2 R0@;";
  fprintf p "ADD R1 R1 #-1@;";
  fprintf p "BRp MULT_LOOP@;";
  fprintf p "@]@;";

  fprintf p "MULT_END: @;@[<v4>    ";
  fprintf p "RET@;";
  fprintf p "@]@;"

(* Fonction de division / modulo *)
let pp_synth_div p () =
  let open Format in
  (* compute R0 / R1 using R2 as dividend and R3 as inverse sign. RET in R0*)
  fprintf p "DIV: @;@[<v4>    ";
  if !Usage.long_jmp then
    fprintf p "ADD %a %a #1 @;" pp_real_reg reg_fun_ret_addr pp_real_reg
      reg_fun_ret_addr;
  fprintf p "AND R3 R3 #0@;";
  fprintf p "AND R2 R2 #0@;";
  fprintf p "ADD R0 R0 #0@;";
  fprintf p "BRzp DIVCONTO@;";
  fprintf p "NOT R3 R3@;";
  fprintf p "NOT R0 R0@;";
  fprintf p "ADD R0 R0 #1@;";
  fprintf p "@]@;";

  fprintf p "DIVCONTO: @;@[<v4>    ";
  fprintf p "ADD R1 R1 #0@;";
  fprintf p "BRn DIVLOOP@;";
  fprintf p "BRz DIVERR@;";
  fprintf p "NOT R3 R3@;";
  fprintf p "NOT R1 R1@;";
  fprintf p "ADD R1 R1 #1@;";
  fprintf p "@]@;";

  fprintf p "DIVLOOP: @;@[<v4>    ";
  fprintf p "ADD R2 R2 #1@;";
  fprintf p "ADD R0 R0 R1@;";
  fprintf p "BRzp DIVLOOP@;";
  fprintf p "ADD R2 R2 #-1@;";
  fprintf p "NOT R1 R1@;";
  fprintf p "ADD R1 R1 #1@;";
  fprintf p "ADD R0 R0 R1@;";
  fprintf p "ADD R3 R3 #0@;";
  fprintf p "BRn DIVRET@;";
  fprintf p "NOT R0 R0@;";
  fprintf p "ADD R0 R0 #1@;";
  fprintf p "NOT R2 R2@;";
  fprintf p "ADD R2 R2 #1@;";
  fprintf p "@]@;";

  fprintf p "DIVRET: @;@[<v4>    ";
  fprintf p "RET@;";
  fprintf p "@]@;";

  fprintf p "DIVERR: @;@[<v4>    ";
  fprintf p "LEA R0 DIVERR_STR@;";
  fprintf p "PUTS@;";
  fprintf p "HALT@;";
  fprintf p "DIVERR_STR .STRINGZ \"Err Div / 0\\n\"@;";
  fprintf p "@]@;"

(* Fonction d'affichage d'entiers *)
let pp_synth_print_int p () =
  (* ; IN R0
     ; R2/R0 nb rem R1 = 10
     ; R4 nb operations *)
  let open Format in
  fprintf p "PRINT_INT: @;@[<v4>    ";
  if !Usage.long_jmp then
    fprintf p "ADD %a %a #1 @;" pp_real_reg reg_fun_ret_addr pp_real_reg
      reg_fun_ret_addr;
  fprintf p "ADD R6 R6 #-2@;";
  fprintf p "STR R7 R6 #1@;";
  fprintf p "STR R4 R6 #0@;";

  fprintf p "AND R1 R1 #0@;";
  fprintf p "ADD R1 R1 #10@;";
  fprintf p "AND R4 R4 #0@;";

  fprintf p "ADD R2 R0 #0@;";
  fprintf p "BRp PRINT_DIV_LOOP@;";
  fprintf p "BRn PRINT_NEG@;";
  fprintf p "LD R0 ZERO@;";
  fprintf p "OUT@;";
  fprintf p "BR PRINT_RET@;";
  fprintf p "@]@;";

  fprintf p "PRINT_NEG: @;@[<v4>    ";
  fprintf p "LD R0 TIRET@;";
  fprintf p "OUT@;";
  fprintf p "NOT R2 R2@;";
  fprintf p "ADD R0 R2 #1@;";
  fprintf p "@]@;";

  fprintf p "PRINT_DIV_LOOP: @;@[<v4>    ";
  fprintf p "ADD R4 R4 #1@;";
  fprintf p "ADD R6 R6 #-1@;";
  fprintf p "%a@;" pp_synth_jsr (label_of_func_name "DIV");
  fprintf p "STR R0 R6 #0@;";
  fprintf p "ADD R0 R2 #0@;";
  fprintf p "BRp PRINT_DIV_LOOP@;";
  fprintf p "LD R1 ZERO@;";
  fprintf p "@]@;";

  fprintf p "PRINT_LOOP: @;@[<v4>    ";
  fprintf p "LDR R0 R6 #0@;";
  fprintf p "ADD R0 R0 R1@;";
  fprintf p "OUT@;";
  fprintf p "ADD R6 R6 #1@;";
  fprintf p "ADD R4 R4 #-1@;";
  fprintf p "BRp PRINT_LOOP@;";
  fprintf p "@]@;";

  fprintf p "PRINT_RET: @;@[<v4>    ";
  fprintf p "LDR R7 R6 #1@;";
  fprintf p "LDR R4 R6 #0@;";
  fprintf p "ADD R6 R6 #2@;";
  fprintf p "RET@;";
  fprintf p "@]@;";

  fprintf p "ZERO: .FILL x30 ; '0'@;";
  fprintf p "TIRET: .FILL x2D ; '-'@;"

(* Fonction de comparaison *)
let pp_synth_cmp p () =
  let open Format in
  (* R0, R1 IN, R3 OUT *)
  (* Return n for LT, z for EQ, p for GT*)
  fprintf p "CMP: @;@[<v4>    ";
  if !Usage.long_jmp then
    fprintf p "ADD %a %a #1 @;" pp_real_reg reg_fun_ret_addr pp_real_reg
      reg_fun_ret_addr;
  fprintf p "AND R3 R3 #0@;";
  fprintf p "ADD R0 R0 #0@;";
  fprintf p "BRp CMPPOS@;";
  fprintf p "ADD R1 R1 #0@;";
  fprintf p "BRnz CMPSUB@;";
  fprintf p "ADD R3 R3 #-1@;";
  (* LT *)
  fprintf p "RET";
  fprintf p "@]@;";

  fprintf p "CMPPOS: @;@[<v4>    ";
  fprintf p "ADD R1 R1 #0@;";
  fprintf p "BRp CMPSUB@;";
  fprintf p "ADD R3 R3 #1@;";
  (* GT *)
  fprintf p "RET";
  fprintf p "@]@;";

  fprintf p "CMPSUB: @;@[<v4>    ";
  fprintf p "NOT R1 R1@;";
  fprintf p "ADD R1 R1 #1@;";
  fprintf p "ADD R3 R0 R1@;";
  fprintf p "RET@;";
  fprintf p "@]@;"

let pp_synth_monop p (rres, op, rr) =
  match op with
  | ADDI i ->
      Format.fprintf p "%a@;" pp_synth_load_loc (0, Reg rr);
      Format.fprintf p "ADD R0 R0 #%d@;" i;
      Format.fprintf p "%a@;" pp_synth_store_loc (Reg rres, 0)
  | MOV ->
      Format.fprintf p "%a@;" pp_synth_load_loc (0, Reg rr);
      Format.fprintf p "%a@;" pp_synth_store_loc (Reg rres, 0)
  | MINUS ->
      Format.fprintf p "%a@;" pp_synth_load_loc (0, Reg rr);
      Format.fprintf p "NOT R0 R0@;ADD R0 R0 #1@;";
      Format.fprintf p "%a@;" pp_synth_store_loc (Reg rres, 0)
  | NOT ->
      Format.fprintf p "%a@;" pp_synth_load_loc (0, Reg rr);
      Format.fprintf p "NOT R0 R0@;";
      Format.fprintf p "%a@;" pp_synth_store_loc (Reg rres, 0)
  | DEREF ->
      Format.fprintf p "%a@;" pp_synth_load_loc (0, Reg rr);
      Format.fprintf p "LDR R0 R0 #0@;";
      Format.fprintf p "%a@;" pp_synth_store_loc (Reg rres, 0)

let rec pp_synth_binop p
    ((rres : asm_cfg_reg), (op : binop), (ra : asm_cfg_reg), (rb : asm_cfg_reg))
    =
  let open Format in
  match op with
  | ADD ->
      fprintf p "%a@;%a@;" pp_synth_load_loc (0, Reg ra) pp_synth_load_loc
        (1, Reg rb);
      fprintf p "ADD R0 R0 R1@;";
      pp_synth_store_loc p (Reg rres, 0)
  | SUB ->
      fprintf p "%a@;%a@;" pp_synth_load_loc (0, Reg ra) pp_synth_load_loc
        (1, Reg rb);
      fprintf p "NOT R1 R1@;";
      fprintf p "ADD R1 R1 #1@;";
      fprintf p "ADD R0 R0 R1@;";
      pp_synth_store_loc p (Reg rres, 0)
  | MUL ->
      fprintf p "%a@;%a@;" pp_synth_load_loc (0, Reg ra) pp_synth_load_loc
        (1, Reg rb);
      fprintf p "%a@;" pp_synth_jsr (label_of_func_name "MULT");
      pp_synth_store_loc p (Reg rres, 2)
  | CMP_EQ | CMP_LE | CMP_LT ->
      fprintf p "%a@;" pp_synth_load_loc (0, Reg ra);
      fprintf p "%a@;" pp_synth_load_loc (1, Reg rb);
      fprintf p "AND R2 R2 #0@;";
      fprintf p "%a@;" pp_synth_jsr (label_of_func_name "CMP");
      fprintf p "BR%s #1@;"
        (match op with
        | CMP_EQ -> "np"
        | CMP_LE -> "p"
        | CMP_LT -> "zp"
        | _ -> failwith "The impossible happened");
      fprintf p "ADD R2 R2 #1@;";
      pp_synth_store_loc p (Reg rres, 2)
  | MOD ->
      fprintf p "%a@;%a@;" pp_synth_load_loc (0, Reg ra) pp_synth_load_loc
        (1, Reg rb);
      fprintf p "%a@;" pp_synth_jsr (label_of_func_name "DIV");
      pp_synth_store_loc p (Reg rres, 0)
  | DIV ->
      fprintf p "%a@;%a@;" pp_synth_load_loc (0, Reg ra) pp_synth_load_loc
        (1, Reg rb);
      fprintf p "%a@;" pp_synth_jsr (label_of_func_name "DIV");
      pp_synth_store_loc p (Reg rres, 2)

let pp_synth_ins p bi =
  let open Format in
  let pp_label = PrettyPrinter.pp_label in
  fprintf p "  ; %a@;" pp_asm_ins bi;
  match bi with
  | Nop -> fprintf p "Nop"
  | LoadSmallCst (r, n) ->
      fprintf p "AND R0 R0 #0@;";
      if n <> 0 then fprintf p "%a@;" pp_big_reg_add (0, 0, n);
      pp_synth_store_loc p (Reg r, 0)
  | Load (r, l) ->
      fprintf p "%a@;%a" pp_synth_load_loc (0, l) pp_synth_store_loc (Reg r, 0)
  | Store (l, r) ->
      fprintf p "%a@;%a" pp_synth_load_loc (0, Reg r) pp_synth_store_loc (l, 0)
  | StoreI (l, r) ->
      fprintf p "%a@;" pp_synth_load_loc (0, l);
      fprintf p "%a@;" pp_synth_load_loc (1, Reg r);
      fprintf p "STR R1 R0 #0"
  | StoreR (a, r) ->
      fprintf p "%a@;" pp_synth_load_loc (0, Reg a);
      fprintf p "%a@;" pp_synth_load_loc (1, Reg r);
      fprintf p "STR R1 R0 #0"
  | Binop (rres, op, ra, rb) -> pp_synth_binop p (rres, op, ra, rb)
  | Addr (rres, l) ->
      fprintf p "%a@;%a" pp_synth_load_addr (0, l) pp_synth_store_loc
        (Reg rres, 0)
  | Monop (rres, op, rr) -> pp_synth_monop p (rres, op, rr)
  | PushArg (r, n) ->
      fprintf p "%a@;" pp_synth_load_loc (0, Reg r);
      pp_synth_str p (0, reg_stack, -3 - n)
  | PushStack (nbpl, nbr) ->
      if !Usage.long_jmp then
        fprintf p "ADD %a %a #1@;" pp_real_reg reg_fun_ret_addr pp_real_reg
          reg_fun_ret_addr;
      (* On alloue de la place pour R4 et R7 *)
      fprintf p "%a@;" pp_big_reg_add (reg_stack, reg_stack, -3);
      (* On copie le précédent cadre et le précédent fun_ret_val*)
      fprintf p "STR %a %a #2@;" pp_real_reg reg_fun_stack pp_real_reg reg_stack;
      fprintf p "STR %a %a #1@;" pp_real_reg reg_fun_ret_addr pp_real_reg
        reg_stack;
      (* On sauvegarde le bas de cadre *)
      fprintf p "ADD %a %a #0@;" pp_real_reg reg_fun_stack pp_real_reg reg_stack;
      (* On alloue de la place pour les paramètres et les locales *)
      pp_big_reg_add p (reg_stack, reg_stack, -nbpl - nbr + 1)
  | Call l -> pp_synth_jsr p l
  | CallR r ->
      fprintf p "%a@;JSRR R0" pp_synth_load_loc (0, Reg r);
      if !Usage.long_jmp then fprintf p "@;NOP"
  | Ret r ->
      fprintf p "%a@;" pp_synth_load_loc (reg_fun_ret_val, Reg r);
      (* On retrouve l'adresse de retour *)
      fprintf p "LDR %a %a #1@;" pp_real_reg reg_fun_ret_addr pp_real_reg
        reg_fun_stack;
      (* On remet le pointeur de pile à la bonne place *)
      fprintf p "ADD %a %a #3@;" pp_real_reg reg_stack pp_real_reg reg_fun_stack;
      (* On reprend la valeur de cadre précédente *)
      fprintf p "LDR %a %a #2@;" pp_real_reg reg_fun_stack pp_real_reg
        reg_fun_stack;
      fprintf p "RET"
  | PopRes r -> pp_synth_store_loc p (Reg r, reg_fun_ret_val)
  | Jmp l ->
      if !Usage.long_jmp then fprintf p "LD R0 #1@;JSRR R0@;.FILL %a" pp_label l
      else fprintf p "BR %a" pp_label l
  | JmpC (r, l1, l2) ->
      fprintf p "%a@;" pp_synth_load_loc (0, Reg r);
      if !Usage.long_jmp then (
        fprintf p "BRz #3@;LD R0 #1@;JSRR R0@;.FILL %a@;" pp_label l1;
        fprintf p "LD R0 #1@;JSRR R0@;.FILL %a" pp_label l2)
      else fprintf p "BRnp %a@;BRz %a" pp_label l1 pp_label l2
  | PrintString s -> fprintf p "%a@;PUTS" pp_synth_load_loc (0, GlobalCst s)

let pp_synth_block p (l, ins) =
  Format.(
    fprintf p "%a: @;@[<v4>    %a@]@;" PrettyPrinter.pp_label l
      (pp_print_list pp_synth_ins)
      ins)

let pp_synth_function p f =
  Format.(
    fprintf p "; %s @;@[<v>%a@]@;" f.name
      (pp_print_list pp_synth_block)
      f.blocks)

let pp_synth p (orig, start_label, pre_asm, cst) =
  let open Format in
  (* Section d'initialisation *)
  fprintf p "@[<v0>.ORIG x%x@;" orig;
  fprintf p "INIT: @;@[<v4>    %a"
    (pp_print_list (fun p i ->
         fprintf p "AND %a %a #0" pp_real_reg i pp_real_reg i))
    (List.init nb_regs Fun.id);
  fprintf p "@;LD %a INIT_STACK@;ADD %a %a #0@;LD %a INIT_GLBCST@;" pp_real_reg
    reg_stack pp_real_reg reg_fun_stack pp_real_reg reg_stack pp_real_reg
    reg_glbcst;
  pp_synth_jsr p start_label;
  fprintf p "@;HALT@]@;";
  fprintf p "INIT_STACK: .FILL x%x@;INIT_GLBCST: .FILL GLBCST@;@;"
    (start_stack - 1);

  (* Section des fonctions *)
  fprintf p "@[<v>%a" (pp_print_list pp_synth_function) pre_asm;

  (* Section des fonctions `builtins` *)
  pp_synth_cmp p ();
  pp_synth_mult p ();
  pp_synth_div p ();
  pp_synth_print_int p ();

  (* Section des constantes *)
  let glbcst =
    GlbCstMap.to_list cst |> List.sort (fun (_, i) (_, j) -> compare i j)
  in
  fprintf p "GLBCST: @;@[<v4>    %a@]@;"
    (pp_print_list (fun p (c, id) ->
         match c with
         | IntCst i -> fprintf p ".FILL #%d" i
         | Global n -> fprintf p ".FILL #0 ; global %s" n
         | StringCst _ -> fprintf p ".FILL %a_VAL" pp_glb_cst_id id
         | FuncLabel l -> fprintf p ".FILL %a" PrettyPrinter.pp_label l))
    glbcst;
  (* Section des chaînes de caractères *)
  fprintf p "; String Values@;@[<v4>    %a@]@;"
    (pp_print_list
       ?pp_sep:(Some (fun _ _ -> ()))
       (fun p (c, id) ->
         match c with
         | IntCst _ | Global _ | FuncLabel _ -> ()
         | StringCst s ->
             fprintf p "%a_VAL: .STRINGZ \"%s\"@;" pp_glb_cst_id id
               (String.escaped s)))
    glbcst;
  fprintf p ".END@;@]"

(* On alloue une place dans la pile pour un registre virtuel *)
let get_reg_pos d r =
  match CfgRegMap.find_opt r.reg_name d.reg_table with
  | Some p -> (d, p)
  | None ->
      let i, d =
        match d.reg_free with
        | [] -> (d.nb_reg_used, { d with nb_reg_used = d.nb_reg_used + 1 })
        | i :: q -> (i, { d with reg_free = q })
      in
      ({ d with reg_table = CfgRegMap.add r.reg_name i d.reg_table }, i)

(* On libère un registre virtuel *)
let free_reg_pos d r =
  match CfgRegMap.find_opt r.reg_name d.reg_table with
  | None -> d
  | Some p ->
      {
        d with
        reg_table = CfgRegMap.remove r.reg_name d.reg_table;
        reg_free = p :: d.reg_free;
      }

(* On récupère et alloue le cas échéant l'identifiant d'une constante *)
let get_gbl_cst d g =
  match GlbCstMap.find_opt g d.glb_cst_table with
  | Some p -> (d, p)
  | None ->
      ( {
          d with
          nb_glb_cst = d.nb_glb_cst + 1;
          glb_cst_table = GlbCstMap.add g d.nb_glb_cst d.glb_cst_table;
        },
        d.nb_glb_cst )

(* On transforme une CFG_location en ASM_location *)
let get_location d (l : location) =
  match l with
  | StringCst s ->
      let d, c = get_gbl_cst d (StringCst s) in
      (d, GlobalCst c)
  | Global s ->
      let d, c = get_gbl_cst d (Global s) in
      (d, GlobalCst c)
  | Param p -> (d, ParamLocal p)
  | Local (l, _) ->
      ( (if d.fun_nb_locals_max <= l then { d with fun_nb_locals_max = l + 1 }
         else d),
        ParamLocal (l + d.fun_nb_args) )
  | FuncLabel f ->
      let d, c =
        get_gbl_cst d (FuncLabel (FunNameMap.find f d.funs_entry_label))
      in
      (d, GlobalCst c)

(* On transforme une instruction cfg en instruction asm *)
let transform_ins d (i : non_final_instruction) =
  match i with
  | Nop -> (d, [ Nop ])
  | Cst (r, n) ->
      let d, rr = get_reg_pos d r in
      if -32 <= n && n < 32 then (d, [ LoadSmallCst (rr, n) ])
      else
        let d, gg = get_gbl_cst d (IntCst n) in
        (d, [ Load (rr, GlobalCst gg) ])
  | Load (r, l) ->
      let d, rr = get_reg_pos d r in
      let d, al = get_location d l in
      (d, [ Load (rr, al) ])
  | Store (l, r) ->
      let d, al = get_location d l in
      let d, rr = get_reg_pos d r in
      (d, [ Store (al, rr) ])
  | StoreI (l, r) ->
      let d, al = get_location d l in
      let d, rr = get_reg_pos d r in
      (d, [ StoreI (al, rr) ])
  | StoreR (a, r) ->
      let d, ra = get_reg_pos d a in
      let d, rr = get_reg_pos d r in
      (d, [ StoreR (ra, rr) ])
  | Addr (r, l) ->
      let d, rr = get_reg_pos d r in
      let d, l' = get_location d l in
      (d, [ Addr (rr, l') ])
  | Monop (res, op, a) ->
      let d, rres = get_reg_pos d res in
      let d, rra = get_reg_pos d a in
      (d, [ Monop (rres, op, rra) ])
  | Binop (res, op, a, b) ->
      let d, rres = get_reg_pos d res in
      let d, rra = get_reg_pos d a in
      let d, rrb = get_reg_pos d b in
      (d, [ Binop (rres, op, rra, rrb) ])
  | PrintString s ->
      let d, g = get_gbl_cst d (StringCst s) in
      (d, [ PrintString g ])
  | PushCallArg (r, n) ->
      let d, rr = get_reg_pos d r in
      (d, [ PushArg (rr, n) ])
  | Call s -> (d, [ Call (FunNameMap.find s d.funs_entry_label) ])
  | CallR r ->
      let d, rr = get_reg_pos d r in
      (d, [ CallR rr ])
  | PopCallRes r ->
      let d, rr = get_reg_pos d r in
      (d, [ PopRes rr ])
  | FreeRegs s -> (RegSet.fold (Fun.flip free_reg_pos) s d, [])

(* On transforme une instruction finale cfg en instruction asm *)
let transform_fins d f =
  match f with
  | Return r ->
      let d, rr = get_reg_pos d r in
      (free_reg_pos d r, Ret rr)
  | Jmp l -> (d, Jmp l)
  | JmpC (r, l1, l2) ->
      let d, rr = get_reg_pos d r in
      (free_reg_pos d r, JmpC (rr, l1, l2))

(* On transforme un bloc cfg en bloc cfg assembleur cf doc *)
let transform_block g d bl =
  let ins, fins = get_block bl g in
  let d, asm_ins = List.fold_left_map transform_ins d ins in
  let d, asm_fins = transform_fins d fins in
  (d, (bl, List.flatten asm_ins @ [ asm_fins ]))

let transform_fun d (f : function_cfg) =
  let cfg_blocks = topo_sort f.entry f.graph in
  let d =
    {
      d with
      fun_nb_args = f.nb_args;
      fun_nb_locals_max = 0;
      reg_table = CfgRegMap.empty;
      nb_reg_used = 0;
      reg_free = [];
    }
  in
  let d, blocks = List.fold_left_map (transform_block f.graph) d cfg_blocks in
  let stl, stins = List.hd blocks in
  let blocks =
    ( stl,
      PushStack (d.fun_nb_args + d.fun_nb_locals_max, d.nb_reg_used) :: stins )
    :: List.tl blocks
  in
  Format.eprintf "SYNT: %s\nRegs used: %d.\nNb blocs: %d\n\n" f.name
    d.nb_reg_used (List.length blocks);
  (d, { name = f.name; blocks })

let synthesis (orig : int) (p : program) =
  let d, (pre_asm : asm_function list) =
    List.fold_left_map transform_fun
      {
        glb_cst_table = GlbCstMap.empty;
        nb_glb_cst = 0;
        fun_nb_args = 0;
        fun_nb_locals_max = 0;
        reg_table = CfgRegMap.empty;
        nb_reg_used = 0;
        reg_free = [];
        funs_entry_label =
          FunNameMap.of_list
            (List.map (fun (f : function_cfg) -> (f.name, f.entry)) p.functions)
          |> FunNameMap.add "print_int" (label_of_func_name "PRINT_INT");
      }
      p.functions
  in
  let d =
    List.fold_left (fun d g -> fst (get_gbl_cst d (Global g))) d p.globals
  in
  (* Format.(eprintf "@[<v>%a@]@\n@\n@." (pp_print_list pp_asm_function) pre_asm); *)
  let dest =
    match
      List.find_opt (fun (f : function_cfg) -> f.name = "main") p.functions
    with
    | None -> raise (Error "No main function !")
    | Some f -> f
  in
  Format.printf "%a" pp_synth (orig, dest.entry, pre_asm, d.glb_cst_table)
