open Cast
open Printf
open Types
open Pretty

exception Error of location * string
exception Anomaly of string

type typ_var =
  | TVvar of typ (* Une variable *)
  | TVfun of typ_fun (* Une fonction globale *)

(* Type pour stocker les déclarations courantes *)
type declaration_tbl = (string, typ_var) Hashtbl.t

(* Renvoie un bel affichage du typ_var *)
let rec get_typ_var_string (t : typ_var) =
  match t with
  | TVvar v -> sprintf "variable de type %s" (get_typ_string v)
  | TVfun f -> sprintf "fonction de type %s" (get_typ_string (Tfun f))

(* Converti un ctyp en typ *)
let rec ctyp_to_typ (t : ctyp) : typ =
  match t with
  | TINT -> Tscalar Tint
  | TPTR p ->
      Tscalar
        (Tptr
           (match ctyp_to_typ p with
           | Tscalar t -> t
           | t ->
               raise
                 (Anomaly
                    (sprintf
                       "ctype_to_typ: Un type scalaire était attendu pour un \
                        type pointeur"))))
  | TFUN (r, args) ->
      Tfun { result = ctyp_to_typ r; args = List.map ctyp_to_typ args }

(* Renvoie le nom de la déclaration d *)
let get_decl_name (d : var_declaration) : string =
  match d with CDECL (_, n, _) -> n | CFUN (_, n, _, _, _) -> n

(* Renvoie la location de la déclaration d *)
let get_decl_location (d : var_declaration) : location =
  match d with CDECL (l, _, _) -> l | CFUN (l, _, _, _, _) -> l

(* Déclare la variable d dans decs *)
let rec declare_var (decs : declaration_tbl) (d : var_declaration) : typ =
  match d with
  | CDECL (_, n, t) ->
      let typ = ctyp_to_typ t in
      Hashtbl.add decs n (TVvar typ);
      typ
  | CFUN (l, _, _, _, _) ->
      raise (Error (l, "Une définition de variable était attendu"))

(* Déclare toutes les variables de var_decls et vérifie les doublons *)
let declare_var_list (decs : declaration_tbl) (var_decls : var_declaration list)
    : typ list =
  (* Dictionnaire pour vérifier les doublons *)
  let bloc_table = Hashtbl.create (List.length var_decls) in
  List.map
    (fun (d : var_declaration) ->
      let name = get_decl_name d in
      if Hashtbl.mem bloc_table name then
        raise (Error (get_decl_location d, sprintf "Redeclaration de %s" name));
      Hashtbl.add bloc_table name ();
      declare_var decs d)
    var_decls

(* Supprime les variables déclarations des  *)
let undeclare_var_list (decs : declaration_tbl)
    (var_decls : var_declaration list) : unit =
  List.iter
    (fun (d : var_declaration) -> Hashtbl.remove decs (get_decl_name d))
    var_decls

(* Compare a et b et renvoie leur type commun s'ils sont égaux *)
let cmp_types (a : typ) (b : typ) : typ option =
  match (a, b) with
  | a, b when a = b -> Some a
  | Tnull, Tscalar (Tptr t) | Tscalar (Tptr t), Tnull -> Some (Tscalar (Tptr t))
  | Tnull, Tfun f | Tfun f, Tnull -> Some (Tfun f)
  | _ -> None

(* Type l'expression (l, e) et renvoie son type *)
let rec type_expr (decs : declaration_tbl) ((l, e) : loc_expr) : typ =
  match e with
  | VAR n -> begin
      (* Recupère et teste le type de la variable n *)
      match Hashtbl.find_opt decs n with
      | None -> raise (Error (l, sprintf "Variable %s non délcarée" n))
      | Some (TVvar t) -> t
      | Some (TVfun f) -> Tfun f
    end
  | CST _ -> Tscalar Tint
  | STRING _ -> Tstring
  | NULL -> Tnull
  | SET_VAR (n, vle) -> begin
      let te = type_expr decs vle in
      (* Recupère et teste le type de la variable n *)
      match Hashtbl.find_opt decs n with
      | None -> raise (Error (l, sprintf "Variable %s non déclarée" n))
      | Some (TVfun _) ->
          raise
            (Error (l, sprintf "Impossible d'affecter %s qui est une fonction" n))
      | Some (TVvar t) -> (
          (* Vérifie la compatibilité du type de la variable et de l'expression *)
          match cmp_types t te with
          | Some t -> t
          | None ->
              raise
                (Error
                   ( l,
                     sprintf "Affectation d'un type %s à %s qui est de type %s"
                       (get_typ_string te) n (get_typ_string t) )))
    end
  | SET_VAL (n, vle) -> begin
      let te = type_expr decs vle in
      (* Récupère le type de la variable n *)
      match Hashtbl.find_opt decs n with
      | None -> raise (Error (l, sprintf "Variable %s non déclarée" n))
      | Some (TVfun _) ->
          raise
            (Error (l, sprintf "Impossible d'affecter %s qui est une fonction" n))
      | Some (TVvar (Tscalar (Tptr ptr))) -> begin
          (* Vérifie la compatibilité du type de la variable et de l'expression *)
          match cmp_types (Tscalar ptr) te with
          | Some t -> t
          | None ->
              raise
                (Error
                   ( l,
                     sprintf "Affectation d'un type %s à %s qui est de type %s"
                       (get_typ_string te) n
                       (get_typ_string (Tscalar ptr)) ))
        end
      | Some (TVvar t) ->
          raise
            (Error
               ( l,
                 sprintf "Affectation d'un type %s à %s qui est de type %s"
                   (get_typ_string t) n (get_typ_string te) ))
    end
  | CALL (n, vle) -> begin
      (* On récupère le type de la fonction n *)
      match Hashtbl.find_opt decs n with
      | None -> raise (Error (l, sprintf "Fonction %s non déclaréé" n))
      | Some (TVvar (Tfun { result; args })) | Some (TVfun { result; args }) ->
          let lepr = List.length vle in
          let largs = List.length args in
          if lepr <> largs then
            raise
              (Error
                 ( l,
                   sprintf
                     "La fonction %s prends %d arguments mais %d ont été donnés"
                     n largs lepr ));
          (* On vérifie le type des arguments *)
          List.iter2
            (fun ((lepr, epr) : loc_expr) (a : typ) ->
              let te = type_expr decs (lepr, epr) in
              match cmp_types te a with
              | Some _ -> ()
              | None ->
                  raise
                    (Error
                       ( lepr,
                         sprintf
                           "Un argument de type %s est attendu mais le type %s \
                            a été donné"
                           (get_typ_string a) (get_typ_string te) )))
            vle args;
          result
      | Some (TVvar t) ->
          raise
            (Error
               ( l,
                 sprintf "La variable %s est de type %s qui n'est pas appelable"
                   n (get_typ_string t) ))
    end
  | OP1 (op, le) -> type_op1 decs l op le
  | OP2 (op, leg, led) -> type_op2 decs l op leg led
  | CMP (_, leg, led) ->
      let tg = type_expr decs leg in
      let td = type_expr decs led in
      begin
        (* Vérifie la compatibilité du type de chaque côté de la comparaison *)
        match cmp_types td tg with
        | Some t -> Tscalar Tint
        | None ->
            raise
              (Error
                 ( l,
                   sprintf "Comparaison de %s et %s qui sont différents"
                     (get_typ_string tg) (get_typ_string td) ))
      end
  | EIF (c, i, e) ->
      (* Comme pour le if, on considère que la condition peut être de n'importe quel type *)
      ignore (type_expr decs c);
      let ti = type_expr decs i in
      let te = type_expr decs e in
      begin
        (* Vérifie la compatibilité du type de chaque côté de la dijonction *)
        match cmp_types ti te with
        | Some t -> t
        | None ->
            raise
              (Error
                 ( l,
                   sprintf
                     "Les deux branches ont pout type %s et %s et ils sont \
                      différents"
                     (get_typ_string ti) (get_typ_string te) ))
      end
  | ESEQ le_list ->
      (* Si c'est le_list est vide, c'est un skip, on peut mettre un type arbitraire *)
      (* On veut juste vérifier le typage de chaque composante et renvoyer le dernier type *)
      List.fold_left (fun _ le -> type_expr decs le) Tnull le_list
(* | _ -> raise (Error (l, "Typage de l'expression non implémenté !")) *)

(* Type une expression opération unaire *)
and type_op1 (decs : declaration_tbl) (loc : location) (op : mon_op)
    (le : loc_expr) : typ =
  (* On vérifie que (l,e) est une variable de type scalaire *)
  let except_scalar_var ((l, e) : loc_expr) : typ_basic =
    match e with
    | VAR n -> begin
        (* On récupère le type de la variable n *)
        match Hashtbl.find decs n with
        | TVvar (Tscalar t) -> t
        | t ->
            raise
              (Error
                 ( l,
                   sprintf
                     "%s est une %s alors que l'opération unaire \"%s\" \
                      attends une variable scalaire"
                     n (get_typ_var_string t) (unop op) ))
      end
    | _ ->
        raise
          (Error
             ( l,
               sprintf
                 "L'opération unaire \"%s\" attends une variable scalaire en \
                  entrée"
                 (unop op) ))
  in
  (* On vérifie que (l,e) est une variable ou un déférencement d'un pointeur *)
  let except_scalar_var_or_val ((l, e) : loc_expr) : typ_basic =
    match e with
    | VAR n -> except_scalar_var (l, VAR n)
    | OP1 (M_DEREF, op_le) -> begin
        (* On récupère et vérifie que c'est bien un pointeur i.e. que *op_le à un sens *)
        match type_expr decs (l, OP1 (M_DEREF, op_le)) with
        | Tscalar t -> t
        | t ->
            raise
              (Error
                 ( l,
                   sprintf
                     "l'opération unaire \"%s\" attends une variable scalaire \
                      et non un type %s"
                     (unop op) (get_typ_string t) ))
      end
    | _ ->
        raise
          (Error
             ( l,
               sprintf
                 "L'opération unaire \"%s\" attends une variable scalaire en \
                  entrée"
                 (unop op) ))
  in
  match op with
  | M_PRE_DEC | M_PRE_INC | M_POST_INC | M_POST_DEC ->
      Tscalar (except_scalar_var_or_val le)
  | M_ADDR -> begin Tscalar (Tptr (except_scalar_var le)) end
  | M_NOT | M_MINUS -> begin
      let t = type_expr decs le in
      match t with
      | Tscalar Tint -> Tscalar Tint
      | _ ->
          raise
            (Error
               ( loc,
                 sprintf "L'opération \"%s\" n'est pas défini pour un pointeur"
                   (unop op) ))
    end
  | M_DEREF -> begin
      let t = type_expr decs le in
      match t with
      | Tscalar (Tptr ptr) -> Tscalar ptr
      | t' ->
          raise
            (Error
               ( loc,
                 sprintf "L'opération \"%s\" n'est pas défini pour %s" (unop op)
                   (get_typ_string t') ))
    end

(* Type une expression opération binaire *)
and type_op2 (decs : declaration_tbl) (loc : location) (op : bin_op)
    (leg : loc_expr) (led : loc_expr) : typ =
  let tg = type_expr decs leg in
  let td = type_expr decs led in
  (* Dijonction de cas sur les différents types de tg (op) td *)
  match (tg, td) with
  (* Cas double entiers *)
  | Tscalar Tint, Tscalar Tint -> Tscalar Tint
  (* Cas entier / pointeur *)
  | Tscalar Tint, Tscalar (Tptr ptr) when op = S_ADD -> Tscalar (Tptr ptr)
  | Tscalar (Tptr ptr), Tscalar Tint when op = S_ADD || op = S_SUB ->
      Tscalar (Tptr ptr)
  (* Cas entier / NULL *)
  | Tscalar Tint, Tnull when op = S_ADD -> Tnull
  | Tnull, Tscalar Tint when op = S_ADD || op = S_SUB -> Tnull
  (* Double (pointeur / NULL) *)
  | Tscalar (Tptr ptr), Tscalar (Tptr ptr') when op = S_SUB && ptr = ptr' ->
      Tscalar Tint
  | Tscalar (Tptr _), Tnull when op = S_SUB -> Tscalar Tint
  | Tnull, Tscalar (Tptr _) when op = S_SUB -> Tscalar Tint
  | Tnull, Tnull when op = S_SUB -> Tscalar Tint
  | tg, td ->
      raise
        (Error
           ( loc,
             sprintf
               "L'opération binaire \"%s\" entre %s et %s n'est pas définie"
               (binop op) (get_typ_string tg) (get_typ_string td) ))

(* Type une commande (l, c) et renvoie s'il elle renvoie bien le type ret dans tous les cas *)
let rec type_loc_code (decs : declaration_tbl) (ret : typ) ((l, c) : loc_code) :
    bool =
  match c with
  | CBLOCK (vl, lc) ->
      ignore (declare_var_list decs vl);
      let res =
        List.fold_left (fun br c -> type_loc_code decs ret c || br) false lc
      in
      undeclare_var_list decs vl;
      res
  | CEXPR e ->
      ignore (type_expr decs e);
      false
  | CIF (cond, i, e) ->
      (* On considère que si c'est un pointeur, on vérifie que c'est pas NULL *)
      ignore (type_expr decs cond);
      let a = type_loc_code decs ret i in
      let b = type_loc_code decs ret e in
      a && b
  | CWHILE (cond, c) ->
      ignore (type_expr decs cond);
      ignore (type_loc_code decs ret c);
      (* la condition peut être toujours fausse *)
      false
  | CRETURN None ->
      (* On considère que c'est un return 0 ou un return NULL *)
      true
  | CRETURN (Some e) -> (
      let te = type_expr decs e in
      (* On vérifie la comptabilité entre la valeur de retour annoncé et celle renvoyé *)
      match cmp_types ret te with
      | None ->
          raise
            (Error
               ( l,
                 sprintf
                   "Un type %s était attendu, mais un type %s a été renvoyé"
                   (get_typ_string ret) (get_typ_string te) ))
      | Some _ -> true)

(* Type une déclaration globale (top-level), variable ou fonction *)
let type_global_var_dec (decs : declaration_tbl) (d : var_declaration) :
    string * typ =
  match d with
  | CDECL (l, n, t) ->
      if Hashtbl.mem decs n then
        raise (Error (l, sprintf "Redeclaration de la variable globale %s" n));
      let vt = ctyp_to_typ t in
      Hashtbl.add decs n (TVvar vt);
      (n, vt)
  | CFUN (l, n, args, r_ctyp, lc) ->
      if Hashtbl.mem decs n then
        raise (Error (l, sprintf "Redeclaration de la fonction %s" n));
      let ret = ctyp_to_typ r_ctyp in
      let t = { result = ret; args = declare_var_list decs args } in
      Hashtbl.add decs n (TVfun t);
      if not (type_loc_code decs ret lc) then
        raise
          (Error
             ( l,
               sprintf
                 "La fonction %s ne renvoie pas un %s dans toutes ses branches \
                  d'exécutions."
                 n (get_typ_string ret) ));
      undeclare_var_list decs args;
      (n, Tfun t)

(* Type un fichier entier à partir de la liste des déclarations globales *)
let check_file (vdecs : var_declaration list) =
  (* associe un nom de variable / fonction à son type *)
  let (decs_types : declaration_tbl) = Hashtbl.create (List.length vdecs + 2) in
  Hashtbl.add decs_types "print_int"
    (TVfun { result = Tscalar Tint; args = [ Tscalar Tint ] });
  Hashtbl.add decs_types "print_string"
    (TVfun { result = Tscalar Tint; args = [ Tstring ] });
  List.map (type_global_var_dec decs_types) vdecs
