(* Main program
 *
 * Calls all steps of the compiler.
 *)

open Format
open Lexing
open Clexer
open Cparser
open Usage
open Cfg

let () =
  let c = open_in file in
  let lb = Lexing.from_channel c in
  try
    let f = Cparser.file Clexer.ctoken lb in
    close_in c;

    if Usage.debug then begin
      let ast_dot_file =
        open_out (Filename.chop_suffix file ".c" ^ "_ast.dot")
      in
      Printf.fprintf ast_dot_file "%s" (Pretty.get_dot_ast f (not !no_pretty));
      close_out ast_dot_file
    end;
    eprintf "Lexing and parsing: done!\n@.";
    if !Usage.parse_only then exit 0;

    let global_decs = Ctyping.check_file f in
    eprintf "Déclarations globales:\n";
    List.iter
      (fun (n, t) -> eprintf "%s: %s\n" n (Pretty.get_typ_string t))
      global_decs;
    eprintf "\n";
    eprintf "Typage: done!\n@.";

    let program = Ast2cfg.transform_program f in

    let program =
      if !Usage.no_analysis_cf then program else Analysis_cf.analyse program
    in
    (* On veut faire l'analyse des initialisation après le Constant Folding, pour ne pas faire de faux
       positif si une branche est prise dans tous les cas cf tst/analysis/cf-loop.c *)
    Analysis_init.analyse program;
    let program =
      if !Usage.no_analysis_filter then program
      else Analysis_filter.analyse program
    in
    begin
      match !Usage.cfg_file with
      | None -> ()
      | Some filename ->
          (if !Usage.cfg_file_simple then Cfg.dump_dot_simple else Cfg.dump_dot)
            filename program
    end;
    if not !Usage.analyse_only then Compile2lc3.synthesis 0x3000 program;
    exit 0
  with
  | Clexer.Lexing_error s ->
      report_loc (lexeme_start_p lb, lexeme_end_p lb);
      eprintf "Lexical error: %s\n@." s;
      exit 1
  | Cparser.Error ->
      report_loc (lexeme_start_p lb, lexeme_end_p lb);
      eprintf "Syntax error\n@.";
      exit 2
  | Ctyping.Error (l, msg) ->
      report_loc l;
      eprintf "Typing error: %s\n@." msg;
      exit 3
  | Ctyping.Anomaly msg ->
      eprintf "Typing Anomaly: %s\n@." msg;
      exit 2
  | Analysis.Anomaly msg ->
      eprintf "Typing Anomaly: %s\n@." msg;
      exit 2
(* | Compile.Anomaly msg ->
   eprintf "Compile anomaly: %s\n@." msg;
   exit 2
*)
(* | e ->
    eprintf "Anomaly: %s\n@." (Printexc.to_string e);
    exit 4 *)
